/** <module> ABS to RBR(abs) Compiler

This module implements the ABS to RBR(abs) compiler, i.e., given an
already loaded ABS program in the loaded_abs_program DB, it produces
an RBR representing such program which is stored in the rbr_abs_db
module (see the documentation there for more details).
  
@author Miky Zamalloa and Antonio Flores
@license GPL
*/

:- module(abs2rbr_compiler,[compile_to_rbr/0, build_method_rule_name/3]).

:- use_module(loaded_abs_db,[def/4,method/4,interface/3,class/3,mainBlock/2,cleanup_ladb/0]).
:- use_module(rbr_abs_db,[add_abs_rule/6,add_call_rel/2,add_rule_field/2,replace_rule_body/3,
			  abs_rule/6, cleanup_rbrdb/0]).
:- use_module(class_invariants,[add_entry_anns/3, add_await_anns/3]).
:- use_module(costabs_output_manager,[info_message/2]).
:- use_module(costabs_params,[costabs_param/2]).
:- use_module(auxiliary_preds,[is_relational_exp/1, is_relational_op/1, is_arithmetic_op/1,
			       negate_relop/2,obtainImplementingInterfaces/2,obtainImplementingClasses/2]).
:- use_module(stdlib(utils),[ut_append/3,ut_atoms_concat/2,
			     ut_number_to_atom/2,ut_flat_list/2,ut_remove/3,
			     ut_member/2,ut_sort_rdup/2,ut_extension/2]).


compile_to_rbr :-
	init_ids,
	compile_defs,
	compile_methods,
	compile_mainBlock,
	(costabs_param(apet_mode,yes) -> generate_await_conts ; true),
	cleanup_abs2rbr.
compile_defs :-
	def(RName,_,DefArgs,DefBody),
	init_s_names,
	take_def_vars(DefArgs,InVars),
	compile_pexp(DefBody,Ret,Body,RName,InVars),
	ut_flat_list(Body,FlatBody),
	new_rule_id(RId),
	add_abs_rule(RName,none,InVars,[Ret],FlatBody,[func,none,RId]),
	fail.
compile_defs.


compile_methods :-
	method(MSig,CN,Anns,Block),
	init_s_names,
	build_method_rule_name(MSig,CN,RName),
	add_entry_anns(CN,RName,Anns),
	Block = block(VarDecls,Stmts),
	take_method_vars(MSig,VarDecls,InArgs,AllVars),
	compile_stmts(Stmts,AllVars,Body,OutVars,RName,CN),
	ut_flat_list(Body,FlatBody),
	new_rule_id(RId),
	add_abs_rule(RName,l(this),InArgs,OutVars,FlatBody,[method,CN,RId]),
	fail.
compile_methods.

compile_mainBlock :-
	mainBlock(VarDecls,Stmts),
	init_s_names,		
	RName = 'main',
	vars_from_var_decls(VarDecls,AllVars),
	InArgs = [],
	compile_stmts(Stmts,AllVars,Body,OutVars,RName,none),
	ut_flat_list(Body,FlatBody),
	new_rule_id(RId),
	add_abs_rule(RName,l(this),InArgs,OutVars,FlatBody,[method,none,RId]),
	fail.
compile_mainBlock.


%% compile_stmts([assignStmt(V,getExp(l(FV)))|Stmts],In,B,Out,RName,CN) :-
%% 	costabs_param(apet_mode,yes), !,
%% 	compile_get_with_cont(V,l(FV),Stmts,In,Out,RName,CN,B).
compile_stmts([assignStmt(V,Exp)|Stmts],In,B,Out,RName,CN) :- !,
	compile_exp(Exp,Aux,BExp,RName,In,Out,CN),
	compile_stmts(Stmts,In,RB,Out,RName,CN),
	(BExp = [Aux=GetOrF] -> B = [V=GetOrF|RB] ; B = [BExp,V=Aux|RB]). % V can be a var or a field(F)

compile_stmts([expressionStmt(Exp)|Stmts],In,B,Out,RName,CN) :- !,
	compile_exp(Exp,_Aux,BExp,RName,In,Out,CN),
	compile_stmts(Stmts,In,RB,Out,RName,CN),
	B = [BExp|RB].

compile_stmts([varDeclStmt(varDecl(_V,_Type,none))|Stmts],In,RB,Out,RName,CN) :- !,
	compile_stmts(Stmts,In,RB,Out,RName,CN).
%% compile_stmts([varDeclStmt(varDecl(V,_Type,getExp(l(FV))))|Stmts],In,B,Out,RName,CN) :-
% 	costabs_param(apet_mode,yes), !,
%% 	compile_get_with_cont(V,l(FV),Stmts,In,Out,RName,CN,B).
compile_stmts([varDeclStmt(varDecl(V,_Type,Exp))|Stmts],In,B,Out,RName,CN) :- !,
	compile_exp(Exp,Aux,BExp,RName,In,Out,CN),
	compile_stmts(Stmts,In,RB,Out,RName,CN),
	(BExp = [Aux=GetOrF] -> B = [V=GetOrF|RB] ; B = [BExp,V=Aux|RB]).

compile_stmts([ifStmt(Guard,Block1,Block2)|Stmts],In,B,Out,RName,CN) :- !,
	build_name(if,IfName),
	add_call_rel(RName,IfName),
	(is_relational_exp(Guard) -> 
	       (Guard =..[RelOp,Exp1,Exp2],
		compile_pexp(Exp1,Aux1,BGuard1,RName,In),
		compile_pexp(Exp2,Aux2,BGuard2,RName,In),
		BGuard = [BGuard1,BGuard2],
		IfArgs = [Aux1,Aux2|In],
		compile_if_branches_rel(RelOp,IfArgs,Block1,Block2,Stmts,IfName,In,Out,CN))
	;
	       (compile_pexp(Guard,Aux,BGuard,RName,In),
		IfArgs = [Aux|In],
		compile_if_branches_norel(IfArgs,Block1,Block2,Stmts,IfName,In,Out,CN))),
	IfCall =..[IfName,l(this),IfArgs,Out],
	GenericCall =..[call,sync,IfCall],
	B = [BGuard,GenericCall].

compile_stmts([whileStmt(Guard,WhileBlock)|Stmts],In,B,Out,RName,CN) :- !,
	build_name(entrywhile,EWName),
	add_call_rel(RName,EWName),
	build_name(while,WhileName),
	add_call_rel(EWName,WhileName),		
	(is_relational_exp(Guard) ->
	       (Guard =..[RelOp,Exp1,Exp2],
		compile_pexp(Exp1,Aux1,BGuard1,RName,In),
		compile_pexp(Exp2,Aux2,BGuard2,RName,In),
		BGuard = [BGuard1,BGuard2],
		WhileArgs = [Aux1,Aux2|In],
		compile_while_branches_rel(RelOp,WhileArgs,WhileName,EWName,OutWhile,WhileBlock,CN,Stmts))
	;
	       (compile_pexp(Guard,Aux,BGuard,RName,In),
		WhileArgs = [Aux|In], 
		compile_while_branches_norel(WhileArgs,WhileName,EWName,OutWhile,WhileBlock,CN,Stmts))),
	EWCall =..[EWName,l(this),In,OutWhile],
	GenericCallEW =..[call,sync,EWCall],
	WhileCall =..[WhileName,l(this),WhileArgs,OutWhile],
	GenericCallW =..[call,sync,WhileCall],
	ut_flat_list([BGuard,GenericCallW],EWBody),
	new_rule_id(RId),
	add_abs_rule(EWName,l(this),In,OutWhile,EWBody,[entrywhile,CN,RId]),
	(costabs_param(apet_mode,no) -> (
	        B = [GenericCallEW|RB],	compile_stmts(Stmts,In,RB,Out,RName,CN))
	;(
	        B = [GenericCallEW], Out = OutWhile)).
compile_stmts([call(sync,EWCall)],_In,[call(sync,EWCall)],_Out,RN,_) :- !,
	% Special case for continuation-based compilation of if-then-else inside whiles
	functor(EWCall,EWName,_),
	add_call_rel(RN,EWName).

compile_stmts([block(VarDecls,Code)|Stmts],In,B,Out,RName,CN) :- !,
	% We only have something in Stmts in the special case of if-else's inside while's
	append(Code,Stmts,Stmts_p),
	vars_from_var_decls(VarDecls,BlockLVs),
	ut_append(In,BlockLVs,In_p),
	compile_stmts(Stmts_p,In_p,B,Out,RName,CN).

%% compile_stmts([returnStmt(getExp(l(FV)))],In,B,[Out],RName,CN) :- 
%% 	costabs_param(apet_mode,yes), !, 
%% 	new_svar(Out),
%% 	compile_get_with_cont(Out,l(FV),[],[Out|In],[Out],RName,CN,B).
compile_stmts([returnStmt(Exp)],In,BExp,[Ret],RName,CN) :- !,
	compile_exp(Exp,Ret,BExp,RName,In,[Ret],CN).

compile_stmts([skipStmt|Stmts],In,RB,Out,RName,CN) :- !, % We ignore skip statements
	compile_stmts(Stmts,In,RB,Out,RName,CN).

compile_stmts([suspendStmt(Id)|Stmts],In,[release(Id)|RB],Out,RName,CN) :- !,
	compile_stmts(Stmts,In,RB,Out,RName,CN).

compile_stmts([awaitStmt(Id,Anns,Guard)|Stmts],In,B,Out,RName,CN) :-
	costabs_param(apet_mode,no), !,
	await_guard(Guard,CGuard),
	compile_stmts(Stmts,In,RB,Out,RName,CN),
	add_await_anns(CN,Id,Anns),
	B = [await(Id,CGuard)|RB].
compile_stmts([awaitStmt(Id,_Anns,Guard)|Stmts],In,B,Out,RName,CN) :-
	costabs_param(apet_mode,yes), !,
	compile_stmts(Stmts,In,RB,Out,RName,CN),
	compile_await_guard(Guard,In,RName,CN,AGCall),
	B = [await(Id,AGCall,In,Out)|RB].

compile_stmts([none|Stmts],In,B,Out,RName,CN) :- !,
	compile_stmts(Stmts,In,B,Out,RName,CN).

compile_stmts([Stmt|_Stmts],_In,_B,_Out,RName,_) :-
	throw(costabs_exception(not_supported_stmt(Stmt,RName))).

compile_stmts([],_,Out,[],RName,_) :-
	var(Out), ut_extension(RName,'.init'), !,
	Out = [].  % This is the end of an init method (we set Out = [])
compile_stmts([],_,[Out = 'Unit'],[Out],_,_) :- % var(Out), !, Out = []. 
	var(Out), !, new_svar(Out). % This is the end of a Unit method
compile_stmts([],_,[],_Out,_,_). % If it is instantiated this is the end of an interm. block


%% compile_get_with_cont(V,l(FV),Stmts,In,Out,RName,CN,B) :-
%% 	compile_stmts(Stmts,In,RB,Out,RName,CN),
%% 	B = [V=get(l(FV),In,Out),RB].


await_guard(claimGuard(Out),Out).
await_guard(expGuard(Exp),Exp).
await_guard(andGuard(G1,G2),Out) :-
	await_guard(G1,CG1),
	await_guard(G2,CG2),
	Out = and(CG1,CG2).

compile_await_guard(AwaitGuard,In,RName,CN,AGCall) :-
	build_name(awguard,AGName),
	add_call_rel(RName,AGName),
	code_for_await_guard(AwaitGuard,In,Code,Out,AGName),
	ut_flat_list(Code,FlatCode),
	new_rule_id(RId),
	add_abs_rule(AGName,l(this),In,[Out],FlatCode,[cont,CN,RId]),
	AGCall =..[AGName,l(this),In,[Out]].
	%GenericCall =..[call,sync,Call].

code_for_await_guard(claimGuard(Fut),_In,[Out = fut_avail(Fut)],Out,_RName) :- new_svar(Out).
code_for_await_guard(expGuard(Exp),In,Code,Out,RName) :-
	compile_pexp(Exp,Out,Code,RName,In).
code_for_await_guard(andGuard(G1,G2),In,Code,Out,RName) :-
	code_for_await_guard(G1,In,Code1,Out1,RName),
	code_for_await_guard(G2,In,Code2,Out2,RName),
	new_svar(Out),
	Code = [Code1,Code2,Out=and(Out1,Out2)].
	

compile_exp(Exp,Ret,B,RN,In,Out,CN) :- nonvar(Exp), compile_iexp(Exp,Ret,B,RN,In,Out,CN), !.
compile_exp(Exp,Ret,B,RN,In,_,_) :- compile_pexp(Exp,Ret,B,RN,In),!.


compile_iexp(MethodCall,Ret,Body,RName,InVars,OutVars,CN) :-
	MethodCall =..[XSync,Exp,InterfacesList,MN,Args],
	member(XSync,[syncCall,asyncCall]), !,
	compile_pexp(Exp,Receiver,BExp,RName,InVars),	
	compile_pexps(Args,Outs,Bodies,RName,InVars),
	(XSync = syncCall -> YSync = sync ; YSync = async), 
	compile_virtualCalls(MN,InterfacesList,Receiver,Outs,RuleCall,Ret,YSync,CN,ImplCs),
	functor(RuleCall,CallRN,_),
	add_call_rel(RName,CallRN),
	((costabs_param(apet_mode,yes), YSync = sync) ->
	        (new_svar(FV),
		 RuleCall =..[CallRN,CallRec,CallIn,[_]],
		 RuleCall_p =..[CallRN,CallRec,CallIn,[FV]],
		 GetAssign = '='(Ret,get(FV,[Ret|InVars],OutVars)),
		 (ImplCs = [CN] -> (compile_await_guard(claimGuard(FV),[FV|InVars],RName,CN,AGCall),
				    Await = await(_,AGCall,[FV|InVars],OutVars),
				    Body = [BExp,Bodies,Call,Await,GetAssign])
		                 ;  Body = [BExp,Bodies,Call,GetAssign]),
		 Call =..[call,async,RuleCall_p])
	;
	        (Call =..[call,YSync,RuleCall],
	         Body = [BExp,Bodies,Call])).

compile_iexp(new(CN,Args,cog),Ret,Body,RName,InVars,_OutVars,_) :- !,
	new_svar(Ret),
	compile_pexps(Args,Outs,Bodies,RName,InVars),
	New =..[newcog,CN],
	build_method_rule_name(methodSig(init,_,_),CN,InitRN),	
	RuleCall =..[InitRN,Ret,Outs,[]],
	Call =..[call,sync,RuleCall],
	Body = [Bodies,Ret=New,Call].
compile_iexp(new(CN,Args,_Cog),Ret,Body,RName,InVars,_OutVars,_) :- !,
	new_svar(Ret),
	compile_pexps(Args,Outs,Bodies,RName,InVars),
	New =..[new,CN],
	build_method_rule_name(methodSig(init,_,_),CN,InitRN),	
	RuleCall =..[InitRN,Ret,Outs,[]],
	Call =..[call,sync,RuleCall],
	Body = [Bodies,Ret=New,Call].
compile_iexp(getExp(l(FV)),Ret,[Ret=GetTerm],_RName,InVars,OutVars,_) :- !,
	% We assume the get can only be applied to a (future) local variable
	new_svar(Ret),
	(costabs_param(apet_mode,yes) -> GetTerm = get(l(FV),InVars,OutVars)
	                               ; GetTerm = get(l(FV))).

compile_pexp(Var,Var,[],_,_) :- var(Var), !.
compile_pexp('_','_',[],_,_) :- !.
compile_pexp(l(Var),l(Var),[],_,_) :- !.
compile_pexp(case(Exp,Branches),Ret,Body,RName,InVars) :- !,
	new_svar(Ret),
	build_name(case,CaseName),
	add_call_rel(RName,CaseName),
	(is_relational_exp(Exp) -> 
	       (Exp =..[RelOp,Exp1,Exp2],
		compile_pexp(Exp1,Aux1,BGuard1,RName,InVars),
		compile_pexp(Exp2,Aux2,BGuard2,RName,InVars),
		BGuard = [BGuard1,BGuard2],
		CaseArgs = [Aux1,Aux2|InVars],
		compile_case_branches_rel(RelOp,CaseArgs,Branches,CaseName))
	;
	       (compile_pexp(Exp,Aux,BGuard,RName,InVars),
		CaseArgs = [Aux|InVars],
		compile_case_branches(CaseArgs,Branches,[],CaseName,InVars))),
	CaseCall =..[CaseName,none,CaseArgs,[Ret]],
	GenericCall =..[call,func,CaseCall],
	Body = [BGuard,GenericCall].
compile_pexp(apply(Call),Ret,Body,RName,InVars) :- !,
	new_svar(Ret),
	Call =..[F|FArgs],
	compile_pexps(FArgs,Outs,Bodies,RName,InVars),
	FuncCall =..[F,none,Outs,[Ret]],
	GenericCall =..[call,func,FuncCall],
	Body = [Bodies,GenericCall],
	add_call_rel(RName,F).
compile_pexp(let((_,Var),Exp1,Exp2),Ret,Body,RName,InVars) :- !,
	new_svar(Ret),
	compile_pexp(Exp1,Out1,Body1,RName,InVars),
	compile_pexp(Exp2,Out2,Body2,RName,InVars),
	Body = [Body1,Var = Out1,Body2,Ret = Out2].
compile_pexp(cons(ConsExp),Ret,Body,RName,InVars) :- !,
	new_svar(Ret),
	ConsExp =..[Cons|ConsExpArgs],
	compile_pexps(ConsExpArgs,Outs,Bodies,RName,InVars),
	BodyConsExp =..[Cons|Outs],
	Body = [Bodies,cons(Ret,BodyConsExp)].
compile_pexp(RelOp,Ret,Body,RName,InVars) :-
	RelOp =..[Op|Exps],
	(is_relational_op(Op) ; Op == neg), !,
	new_svar(Ret),
	compile_pexps(Exps,Outs,Bodies,RName,InVars),
	ArithCall =..[Op,none,Outs,[Ret]],
	GenericCall =..[call,func,ArithCall],
	Body = [Bodies,GenericCall],
	add_call_rel(RName,Op).
compile_pexp(ArithOp,Ret,Body,RName,InVars) :-
	ArithOp =..[Op|Exps], % Op can be unary (minus) or binary
	is_arithmetic_op(Op), !,
	new_svar(Ret),
	compile_pexps(Exps,Outs,Bodies,RName,InVars),
	ArithExp =..[Op|Outs],
	Body = [Bodies,Ret=ArithExp].
compile_pexp(field(F),field(F),[],RName,_) :-
	costabs_param(field_vars_in_rbr,yes), !,
	add_rule_field(RName,F).
compile_pexp(field(F),Ret,[Ret = field(F)],_,RName) :-
	costabs_param(field_vars_in_rbr,no), !,
	new_svar(Ret), add_rule_field(RName,F).
compile_pexp(builtin,Ret,[native],_,_) :- !, new_svar(Ret).
compile_pexp(Literal,Ret,[Ret = Literal],_,_) :- % For numbers, null and strings
	new_svar(Ret).


compile_pexps([],[],[],_,_).
compile_pexps([E|REs],[Ret|RRets],[B|RBs],RName,InVars) :-
	compile_pexp(E,Ret,B,RName,InVars),
	compile_pexps(REs,RRets,RBs,RName,InVars).


compile_case_branches_rel(_,_,[],_).
compile_case_branches_rel(RelOp,[Aux1,Aux2|In],[branch(TOrF,FExp)|Branches],CaseName) :-
	compile_pexp(FExp,Ret,FExpBody,CaseName,In),
	(TOrF == 'True' -> CaseGuard =..[RelOp,Aux1,Aux2]
	                 ;(negate_relop(RelOp,NRelOp),
			   CaseGuard =..[NRelOp,Aux1,Aux2])),
	CaseBody = [CaseGuard,FExpBody],
	ut_flat_list(CaseBody,FlatBody),
	new_rule_id(RId),
	add_abs_rule(CaseName,none,[Aux1,Aux2|In],[Ret],FlatBody,[case,none,RId]),
	compile_case_branches_rel(RelOp,[Aux1,Aux2|In],Branches,CaseName).

compile_case_branches(_,[],_,_,_) :- !.
compile_case_branches(Args,[branch(Patt,FExp)|RBranches],PrevPatts,Name,InVars) :-
	init_s_names,	
	Args = [V|_],
	build_nomatch_list(PrevPatts,V,NoMatchL),
	vars_from_pattern([Patt],PattVars),
	ut_append(InVars,PattVars,InVars_p),
	compile_pexp(FExp,Ret,FExpBody,Name,InVars_p),
	Body = [NoMatchL,match(V,Patt),FExpBody],
	ut_flat_list(Body,FlatBody),
	new_rule_id(RId),
	add_abs_rule(Name,none,Args,[Ret],FlatBody,[case,none,RId]),
	ut_append(PrevPatts,[Patt],PrevPatts_p),
	compile_case_branches(Args,RBranches,PrevPatts_p,Name,InVars).

compile_if_branches_rel(RelOp,[Aux1,Aux2|In],Block1,Block2,Stmts,IfName,In,Out,CN) :-
	% First we compile the continuation in rule ContName
	% init_s_names, we cannot since the Aux var name could be reused
	build_name(cont,ContName),
	add_call_rel(IfName,ContName),	
	compile_stmts(Stmts,In,ContBody,Out,ContName,CN),
	ut_flat_list(ContBody,FlatContBody),
	new_rule_id(ContId),
	add_abs_rule(ContName,l(this),In,Out,FlatContBody,[cont,CN,ContId]),

	% Then we compile block1 and block2 (both call ContName rule)
	ContCall =..[ContName,l(this),In,Out],
	GenericCall =..[call,sync,ContCall],

	% Here we assume there cannot be a return inside the if nor the else
	% Thus, we discard their outs
	compile_stmts([Block1],In,Body1,Out,IfName,CN),
	ut_flat_list([Body1,GenericCall],FlatBody1),
	IfGuard =..[RelOp,Aux1,Aux2],
	new_rule_id(IfId),
	add_abs_rule(IfName,l(this),[Aux1,Aux2|In],Out,[IfGuard|FlatBody1],[if,CN,IfId]),

	compile_stmts([Block2],In,Body2,Out,IfName,CN),
	ut_flat_list([Body2,GenericCall],FlatBody2),
	negate_relop(RelOp,NRelOp),
	ElseGuard =..[NRelOp,Aux1,Aux2],
	new_rule_id(ElseId),
	add_abs_rule(IfName,l(this),[Aux1,Aux2|In],Out,[ElseGuard|FlatBody2],[if,CN,ElseId]).

compile_if_branches_norel([Aux|In],Block1,Block2,Stmts,IfName,In,Out,CN) :-
	% First we compile the continuation in rule ContName
	build_name(cont,ContName),
	add_call_rel(IfName,ContName),	
	compile_stmts(Stmts,In,ContBody,Out,ContName,CN),
	ut_flat_list(ContBody,FlatContBody),
	new_rule_id(ContId),
	add_abs_rule(ContName,l(this),In,Out,FlatContBody,[cont,CN,ContId]),

	% Then we compile block1 and block2 (both call ContName rule)
	ContCall =..[ContName,l(this),In,Out],
	GenericCall =..[call,sync,ContCall],

	% Here we assume there cannot be a return inside the if nor the else
	% Thus, we discard their outs
	compile_stmts([Block1],In,Body1,Out,IfName,CN),
	ut_flat_list([Body1,GenericCall],FlatBody1),
	new_rule_id(IfId),
	add_abs_rule(IfName,l(this),[Aux|In],Out,[eq(Aux,'True')|FlatBody1],[if,CN,IfId]),

	compile_stmts([Block2],In,Body2,Out,IfName,CN),
	ut_flat_list([Body2,GenericCall],FlatBody2),
	new_rule_id(ElseId),
	add_abs_rule(IfName,l(this),[Aux|In],Out,[eq(Aux,'False')|FlatBody2],[if,CN,ElseId]).

compile_while_branches_rel(RelOp,[Aux1,Aux2|In],WName,EWName,WhileOuts,Block,CN,Stmts) :-
	% We build the call to the entrywhile rule and append it to the code of the block
	(costabs_param(apet_mode,no) -> compute_while_outs(Block,WhileOuts)
	                              ; WhileOuts = Out),
	WhileCall =..[EWName,l(this),In,WhileOuts],
	GenericCall =..[call,sync,WhileCall], 
	compile_stmts([Block,GenericCall],In,WBody,WhileOuts,WName,CN),
	ut_flat_list(WBody,FlatWBody),
	
	% Then we create the two while rules
	negate_relop(RelOp,NRelOp),
	ExitGuard =..[NRelOp,Aux1,Aux2], 
	new_rule_id(WId1), new_rule_id(WId2),
	(costabs_param(apet_mode,yes) -> ( % In apet mode the continuation goes in the while base case
	        compile_stmts(Stmts,In,ContBody,Out,EWName,CN),
		ut_flat_list(ContBody,FlatContB),
		add_abs_rule(WName,l(this),[Aux1,Aux2|In],WhileOuts,[ExitGuard|FlatContB],[while,CN,WId1]))
	;
	        add_abs_rule(WName,l(this),[Aux1,Aux2|In],WhileOuts,[ExitGuard],[while,CN,WId1])),
	WhileGuard =..[RelOp,Aux1,Aux2],
	add_abs_rule(WName,l(this),[Aux1,Aux2|In],WhileOuts,[WhileGuard|FlatWBody],[while,CN,WId2]).

compile_while_branches_norel([Aux|In],WName,EWName,WhileOuts,Block,CN,Stmts) :-
	% We build the call to the entrywhile rule and append it to the code of the block
	(costabs_param(apet_mode,no) -> compute_while_outs(Block,WhileOuts)
	                              ; WhileOuts = Out),
	WhileCall =..[EWName,l(this),In,WhileOuts],
	GenericCall =..[call,sync,WhileCall],
	compile_stmts([Block,GenericCall],In,WBody,WhileOuts,WName,CN),
	ut_flat_list(WBody,FlatWBody),

	% Then we create the two while rules
	new_rule_id(WId1), new_rule_id(WId2),
	(costabs_param(apet_mode,yes) -> ( % In apet mode the continuation goes in the while base case
	        compile_stmts(Stmts,In,ContBody,Out,EWName,CN),
		ut_flat_list(ContBody,FlatContB),
		add_abs_rule(WName,l(this),[Aux|In],WhileOuts,[eq(Aux,'False')|FlatContB],[while,CN,WId1]))
	;
	        add_abs_rule(WName,l(this),[Aux|In],WhileOuts,[eq(Aux,'False')],[while,CN,WId1])),
	add_abs_rule(WName,l(this),[Aux|In],WhileOuts,[eq(Aux,'True')|FlatWBody],[while,CN,WId2]).


compile_virtualCalls(MN,[Interface],Receiver,Outs,RuleCall,Out,YSync,CN,ClassesList) :-
	obtainImplementingInterfaces([Interface],Interfaces),
	obtainImplementingClasses(Interfaces,ClassesList),
	% getMethodReturnType(ClassesList,MN,RetType),
	% (RetType == 'Unit' -> OutL = []; (new_svar(Out), OutL = [Out])),
	new_svar(Out),
	OutL = [Out],
	(ClassesList = [Class] ->
		build_method_rule_name(methodSig(MN,_,_),Class,RuleName)
	;
		create_virtualRule(ClassesList,MN,Receiver,Outs,OutL,RuleName,YSync,CN)
	),
	RuleCall =..[RuleName,Receiver,Outs,OutL], !.

compile_virtualCalls(_MN,Interfaces,_Receiver,_Outs,_RuleCall,_YSync,_,_) :-
	throw(ambiguous_method_call(Interfaces)).

create_virtualRule([],_MN,_Receiver,_InVars,_OutL,RuleName,_YSync,_) :-
	build_name(virtual,RuleName).
	
create_virtualRule([Class|ClassList],MN,Receiver,InVars,OutL,RuleName,YSync,CN) :-
	create_virtualRule(ClassList,MN,Receiver,InVars,OutL,RuleName,YSync,CN),
	build_method_rule_name(methodSig(MN,_,_),Class,CallName),
	CallPred =..[CallName,Receiver,InVars,OutL],
	Call =..[call,YSync,CallPred],
	Body = [type(Receiver,Class),Call],
	new_rule_id(RId),
	add_abs_rule(RuleName,Receiver,InVars,OutL,Body,[virtual,CN,RId]).

% It obtains the return type of first method that matches with the name in the given classes
% TODO check if more than one method can match and how to decide in that case
getMethodReturnType([],MN,_RetType):-
	throw(virtualCallsError('the method is not implemented by the inferred classes',MN)).
getMethodReturnType([Class|_IList],MN,RetType):-
	method(methodSig(MN,RetType,_),Class,_,_),!.
getMethodReturnType([_Class|ClassList],MN,RetType):-
	getMethodReturnType(ClassList,MN,RetType).



%%%%%%%%%%%%%%%%%% Generation of await and get continuations %%%%%%%%%%%%%%%%%%

:- dynamic pending/1.

generate_await_conts :-
	retractall(pending(_)),
	% I do a findall since rbr_rules are added dynamically
	findall((RN,RId),abs_rule(RN,_,_,_,_,[_,_,RId]),RuleIds),
	member((RN,RId),RuleIds),
	abs_rule(RN,_This,_In,_Out,Body,[_,CN,RId]),
	process_body(Body,Body_p,Changed,RN,CN),
	(Changed == yes -> replace_rule_body(RN,RId,Body_p) ; true),
	fail.
generate_await_conts :- process_pending.

process_pending :- 
	retract(pending(RN)), !, % In this case there can only be one rule with name RN
	abs_rule(RN,_This,_In,_Out,Body,[_,CN,RId]), 
	process_body(Body,Body_p,Changed,RN,CN),
	(Changed == yes -> replace_rule_body(RN,RId,Body_p) ; true),
	process_pending.
process_pending.


process_body([],[],no,_,_).
process_body([await(Id,G,In,Out)|RB],B_p,yes,RN,CN) :- !,
 	build_name(cont,ContName),
	add_call_rel(RN,ContName),
	new_rule_id(RId),
	add_abs_rule(ContName,l(this),In,Out,RB,[cont,CN,RId]),
	ContCall =..[ContName,l(this),In,Out],
	GenericCall =..[call,sync,ContCall],
	assertz(pending(ContName)),
	B_p = [await(Id,G),GenericCall].
process_body([V=get(FV,In,Out)|RB],B_p,yes,RN,CN) :- !,
 	build_name(cont,ContName),
 	add_call_rel(RN,ContName),	
 	new_rule_id(RId),
	%(V = s(_) -> In_p = [V|In] ; In_p = In),
 	add_abs_rule(ContName,l(this),In,Out,RB,[cont,CN,RId]),
 	ContCall =..[ContName,l(this),In,Out],
 	GenericCall =..[call,sync,ContCall],
	assertz(pending(ContName)),
 	B_p = [V=get(FV),GenericCall].
process_body([B|RB],[B|RB_p],Changed,RN,CN) :-
	process_body(RB,RB_p,Changed,RN,CN).




%%%%%%%%%%%%%%%%%%%%%%%% Auxiliary predicates %%%%%%%%%%%%%%%%%%%%%%


:- dynamic id/1. % Stores the current Id for "case", "if", "while", "cont",
                 % and "awguard" rules, as well as for stack variables

build_name(T,Name) :-
	IdTerm =..[T,IdN],
	id(IdTerm),
	inc_id(T),
	ut_number_to_atom(IdN,IdA),
	ut_atoms_concat([T,'_',IdA],Name).

new_svar(SVar) :-
	SVar = s(_),
	id(SVar),
	inc_id(s).

new_rule_id(RId) :-
	id(rule(RId)),
	inc_id(rule).

init_s_names :-
	retractall(id(s(_))),
	assertz(id(s(0))).

init_ids :-
	retractall(id(_)),
	assertz(id(case(0))),
	assertz(id(if(0))),
	assertz(id(cont(0))),
	assertz(id(while(0))),
	assertz(id(entrywhile(0))),
	assertz(id(s(0))),
	assertz(id(virtual(0))),
	assertz(id(awguard(0))),
	assertz(id(await(0))),
	assertz(id(rule(0))).

init_id(F) :-
	T =..[F,_],
	retractall(id(T)),
	T_p =..[F,0],		
	assertz(id(T_p)).

inc_id(F) :-
	T =..[F,N],
	retract(id(T)),
	N_p is N+1,
	T_p =..[F,N_p],	
	assertz(id(T_p)).


compute_while_outs(BlockOrStmt,Outs_p) :-
	(BlockOrStmt = block(_,Stmts) -> true
	                               ; Stmts = [BlockOrStmt]),
	compute_stmts_outs(Stmts,Outs,[]),
	ut_sort_rdup(Outs,Outs_p).

compute_stmts_outs([],Ac,Ac).
compute_stmts_outs([assignStmt(l(V),_)|Stmts],Outs,Ac) :- !,
	compute_stmts_outs(Stmts,Outs,[l(V)|Ac]).
compute_stmts_outs([varDeclStmt(varDecl(l(V),_,Exp))|Stmts],Outs,Ac) :-
	Exp \= none, !,
	compute_stmts_outs(Stmts,Outs,[l(V)|Ac]).
compute_stmts_outs([ifStmt(_,B1,B2)|Stmts],Outs,Ac) :- !,
	compute_stmts_outs([B1,B2],Outs_p,[]),
	ut_append(Outs_p,Ac,Ac_p),
	compute_stmts_outs(Stmts,Outs,Ac_p).
compute_stmts_outs([whileStmt(_,B)|Stmts],Outs,Ac) :- !,
	compute_stmts_outs([B],Outs_p,[]),
	ut_append(Outs_p,Ac,Ac_p),
	compute_stmts_outs(Stmts,Outs,Ac_p).
compute_stmts_outs([block(_,B)|Stmts],Outs,Ac) :- !,
	compute_stmts_outs(B,Outs_p,[]),
	ut_append(Outs_p,Ac,Ac_p),
	compute_stmts_outs(Stmts,Outs,Ac_p).
compute_stmts_outs(none,Ac,Ac) :- !.
compute_stmts_outs([_|Stmts],Outs,Ac) :-
	compute_stmts_outs(Stmts,Outs,Ac).


%% Predicates for computing the output vars of a block given the body
%  restricted to the input vars
compute_outWhileVars(Body,In,OutWhile):-
	compute_outWhileVars_(Body,In,[],OutWhile).

compute_outWhileVars_(_Body,[],Accum,Accum):-!.%if there are no more input vars we have them all
compute_outWhileVars_([],_In,Accum,Accum).
compute_outWhileVars_([Var1=_Var2|Body],In,Accum,OutWhile):-!,
	filter_element(Var1,In,In2,Accum,Accum2),
	compute_outWhileVars_(Body,In2,Accum2,OutWhile).
compute_outWhileVars_([call(_Type,Call)|Body],In,Accum,OutWhile):-!,
	Call=..[_Name,_This,_Args,Outs],
	filter_list(Outs,In,In2,Accum,Accum2),
	compute_outWhileVars_(Body,In2,Accum2,OutWhile).
compute_outWhileVars_([_|Body],In,Accum,OutWhile):-
	compute_outWhileVars_(Body,In,Accum,OutWhile).

%% It applies filter_element to all the element of a given list
filter_list([],List,List,Accum,Accum).
filter_list([Element|ListElem],List,ListOut,Accum,AccumOut):-
	filter_element(Element,List,ListAux,Accum,AccumAux),
	filter_list(ListElem,ListAux,ListOut,AccumAux,AccumOut).

%% If the given element is in the list, it is removed form the list
%  and added to the accum list
filter_element(Element,List,ListOut,Accum,AccumOut):-
	ut_member(Element,List),!,
	ut_remove(List, Element, ListOut),
	AccumOut=[Element|Accum].
filter_element(_Element,List,List,Accum,Accum).


build_method_rule_name(methodSig(Name,_,_),Class,RN) :-
	ut_atoms_concat([Class,'.',Name],RN).

take_def_vars([],[]).
take_def_vars([(_,Var)|RParams],[Var|RVars]) :-
	take_def_vars(RParams,RVars).

take_method_vars(MSig,VarDecls,AVs,AllVars) :-
	vars_from_var_decls(VarDecls,LVs),
	MSig = methodSig(_,_,Params),
	take_def_vars(Params,AVs),
	ut_append(AVs,LVs,AllVars).

vars_from_var_decls([],[]).
vars_from_var_decls([varDecl(V,_,_)|VarDecls],[V|Vs]) :-
	vars_from_var_decls(VarDecls,Vs).

vars_from_pattern([],[]).
vars_from_pattern([V|Ts],[V|Vs]) :-
	var(V), !,
	vars_from_pattern(Ts,Vs).
vars_from_pattern([l(V)|Ts],[l(V)|Vs]) :- !,
	vars_from_pattern(Ts,Vs).
vars_from_pattern([T|Ts],Vs) :-
	compound(T), !,
	T =..[_|Args],
	vars_from_pattern(Args,TVs),
	vars_from_pattern(Ts,TsVs),
	ut_append(TVs,TsVs,Vs).
vars_from_pattern([_|Ts],Vs) :-
	vars_from_pattern(Ts,Vs).


build_nomatch_list([],_,[]).
build_nomatch_list([P|Ps],V,[nomatch(V,P)|RNoMatchL]) :-
	build_nomatch_list(Ps,V,RNoMatchL).


cleanup_abs2rbr :-
	retractall(id(_)).
