/** <module> ABS programs loader

This module provides predicate load_abs_program/1, which given a file
name with an ABS program in it, calls the Prolog backend, reads the
generated file and asserts the Prolog terms into the loaded_abs_db
DB. It actually asserts only those terms corresponding to reachable
code from the provided entries.  The Prolog terms are of the following
forms:

* For the compiler directives:
	module/1 specifies the loaded module

* For functional ABS: data/2, type/2 and def/4, represents, resp.,
data type declarations, type synonyms and function definitions.

* For imperative ABS: 
	interface/3 represents interface definitions
	class/3 represents class definitions
	method/4  represents method implementations
  
@author Miky Zamalloa and Antonio Flores
@license GPL
*/

:- module(abs_program_loader,[load_abs_program/2]).


:- use_module(stdlib(utils),[ut_member/2,ut_atoms_concat/2,ut_atom_concat/3,ut_file_exists/1,
			     ut_append/3,ut_atom_codes/2]).
:- use_module(costabs_params,[costabs_param/2]).
:- use_module(loaded_abs_db,[add_module/1,
			     add_type/2, add_data/2, add_def/4,
			     add_interface/3, add_class/3, add_method/4, add_mainBlock/2,
			     def/4, method/4, mainBlock/2]).
:- use_module(costabs_output_manager,[info_message/2]).
:- use_module(stdlib(shell_utilities),[get_env_atom/2,make_dirpath/1]).
:- use_module(abs2rbr_compiler,[build_method_rule_name/3]).
:- use_module(auxiliary_preds,[is_relational_op/1, is_arithmetic_op/1,obtainImplementingInterfaces/2,
			       obtainImplementingClasses/2]).
:- use_module(costabs(costabs_settings),[costabs_path/1]).


load_abs_program(AbsFNs,Entries) :-
	(AbsFNs == [] -> true
	               ; call_PrologBackend(AbsFNs)),
	(ut_file_exists('/tmp/costabs/absPL/abs.pl') ->
		true
	;
		throw(costabs_exception(abs_pl_file_not_found('/tmp/costabs/absPL/abs.pl')))),
	open('/tmp/costabs/absPL/abs.pl',read,SIn),
	read_and_store_terms(SIn),
	close(SIn),
	filter_unreachable_code(Entries).


		
% Looking in examples folder
call_PrologBackend(AbsFNs) :-
	AbsFNs = [AbsFN], % By now we assume there is only one abs file
	% Check if the file exists
	costabs_path(CostabsPath),
	ut_atoms_concat([CostabsPath,'/examples/',AbsFN],AbsPath),
	ut_file_exists(AbsPath),
	(costabs_param(stdlib,no) -> StdLib = ' -nostdlib ' ; StdLib = ' '),
	ut_atoms_concat([CostabsPath,'bin/generateProlog -d /tmp/costabs/absPL',StdLib,AbsPath],Command),
	make_dirpath('/tmp/costabs/absPL'), !,
	shell(Command).

% Looking for full path
call_PrologBackend(AbsPaths) :-
	AbsPaths = [AbsPath], % By now we assume there is only one abs file
	% Check if the file exists
	ut_file_exists(AbsPath),
	(costabs_param(stdlib,no) -> StdLib = ' -nostdlib ' ; StdLib = ' '),
	costabs_path(CostabsPath),
	ut_atoms_concat([CostabsPath,'bin/generateProlog -d /tmp/costabs/absPL',StdLib,AbsPath],Command),
	make_dirpath('/tmp/costabs/absPL'), !,
	shell(Command).

% Else throw exception
call_PrologBackend(AbsFNs) :-
	AbsFNs = [AbsFN], % By now we assume there is only one abs file
	throw(costabs_exception(wrong_abs_file(AbsFN))).


read_and_store_terms(SIn) :-
	read(SIn,Term),
	(Term == end_of_file ->
	    true
	;(
	    store_term(Term),
	    read_and_store_terms(SIn))).

store_term(module(ModuleName)) :- add_module(ModuleName), !.

store_term(type(Name,OriginalName)) :- add_type(Name,OriginalName), !.
store_term(data(Name,Constructors)) :- add_data(Name,Constructors), !.
store_term(def(Name,ReturnType,Arguments,Implementation)) :- 
	add_def(Name,ReturnType,Arguments,Implementation), !.

store_term(interface(Name,Extends,Methods)) :- add_interface(Name,Extends,Methods), !.
store_term(class(Name,ImplementedInterfaces,Fields)):-
	add_class(Name,ImplementedInterfaces,Fields),!.
store_term(methodImpl(MSig,Class,Anns,Block)) :- add_method(MSig,Class,Anns,Block), !.
store_term(mainBlock(VarList,Stmts)) :- add_mainBlock(VarList,Stmts), !.
store_term(Other) :- info_message(not_supported_term,[Other]).




%% Filtering of entries: We filter away all the methods/functions
%  which are unreachable from the entries

:- dynamic reachable_method/2, reachable_def/1, reachable_mainBlock/0.


% Special cases of entries
filter_unreachable_code(all) :- !.
filter_unreachable_code(all-defs) :- !,
	findall(Def,def(Def,_RetType,_DefArgs,_DefBody),AllDefs),
	filter_unreachable_code(AllDefs).
filter_unreachable_code(all-methods) :- !,
	findall(MethodId,
	(
		method(Msig,ClassName,_Anns,_Block),
		build_method_rule_name(Msig,ClassName,MethodId)
	),
	AllMethods),
	% We add the main block if it exists
	(mainBlock(_,_)->
		AllMehodsWithMB=[main|AllMethods]
	;
		AllMehodsWithMB=AllMethods
	),
	filter_unreachable_code(AllMehodsWithMB).

				% General case where we have a list of entries that are going to be analyzed
filter_unreachable_code(Entries):-
	cleanup_reachability,
	compute_reachability(Entries),
	cond_remove_mainBlock,
	remove_unreachable_methods,
	remove_unreachable_defs.

cleanup_reachability :-
	retractall(reachable_method(_,_)),
	retractall(reachable_def(_)),
	retractall(reachable_mainBlock).

% Remove unreachable code
cond_remove_mainBlock :- reachable_mainBlock.
cond_remove_mainBlock :- retractall(mainBlock(_,_)).

remove_unreachable_methods :-
	method(methodSig(MethodName,RetType,Params),ClassName,_,_),
	(reachable_method(MethodName,ClassName) ->
		true
	;
		retract(method(methodSig(MethodName,RetType,Params),ClassName,_,_))
	), fail.
remove_unreachable_methods.

remove_unreachable_defs :-
	def(Def,RetType,DefArgs,DefBody),
	(reachable_def(Def) ->
		true
	;
		retract(def(Def,RetType,DefArgs,DefBody))
	), fail.
remove_unreachable_defs.


compute_reachability([]).
compute_reachability([main|Rest]) :- !,
	compute_mainBlock_reachability,
	compute_reachability(Rest).
compute_reachability([Entry|Rest]) :-
	is_method_entry(Entry,ClassName,MethodName), !,
	compute_method_reachability(MethodName,ClassName),
	compute_reachability(Rest).
compute_reachability([Entry|Rest]) :-
	compute_def_reachability(Entry),
	compute_reachability(Rest).

% main block
compute_mainBlock_reachability :-
	mainBlock(_,Stmts), !,
	assertz(reachable_mainBlock),
	compute_stmts_reachability(Stmts).
compute_mainBlock_reachability :-
	throw(costabs_exception(inexistent_entry(main_block))).

% methods
compute_method_reachability(MethodName,ClassName) :-
	reachable_method(MethodName,ClassName), !.
compute_method_reachability(MethodName,ClassName) :-
	method(methodSig(MethodName,_RetType,_Params),ClassName,_,Block), !,
	assertz(reachable_method(MethodName,ClassName)),
	Block = block(_VarDecls,Stmts),
	compute_stmts_reachability(Stmts).
compute_method_reachability(MethodName,ClassName) :-
	build_method_rule_name(methodSig(MethodName,_RetType,_Params),ClassName,Entry),
	throw(costabs_exception(inexistent_entry(Entry))).

compute_virtual_methods_reachability([],_MN).
compute_virtual_methods_reachability([Class|List],MN) :-
	compute_method_reachability(MN,Class),
	compute_virtual_methods_reachability(List,MN).

% defs
compute_def_reachability(Def) :-
	reachable_def(Def), !.
compute_def_reachability(Def) :-
	def(Def,_,_DefArgs,DefBody), !,
	assertz(reachable_def(Def)),
	compute_exp_reachability(DefBody).
compute_def_reachability(Entry) :-
	throw(costabs_exception(inexistent_entry(Entry))).

	
% Stmts
compute_stmts_reachability([]).
compute_stmts_reachability([assignStmt(_V,Exp)|Stmts]) :- !,
	compute_exp_reachability(Exp),
	compute_stmts_reachability(Stmts).
compute_stmts_reachability([expressionStmt(Exp)|Stmts]) :- !,
	compute_exp_reachability(Exp),
	compute_stmts_reachability(Stmts).
compute_stmts_reachability([varDeclStmt(varDecl(_V,_Type,Exp))|Stmts]) :- !,
	compute_exp_reachability(Exp),
	compute_stmts_reachability(Stmts).
compute_stmts_reachability([ifStmt(Guard,Block1,Block2)|Stmts]) :- !,
	compute_exp_reachability(Guard),
	compute_stmts_reachability([Block1,Block2|Stmts]).
compute_stmts_reachability([whileStmt(Guard,WhileBlock)|Stmts]) :- !,
	compute_exp_reachability(Guard),
	compute_stmts_reachability([WhileBlock|Stmts]).
compute_stmts_reachability([block(_,Stmts1)|Stmts]) :- !,
	compute_stmts_reachability(Stmts1),
	compute_stmts_reachability(Stmts).
compute_stmts_reachability([awaitStmt(_,_,Guard)|Stmts]) :- !,
	compute_guard_reachability(Guard),
	compute_stmts_reachability(Stmts).
compute_stmts_reachability([returnStmt(Exp)|Stmts]) :- !,
	compute_exp_reachability(Exp),
	compute_stmts_reachability(Stmts).
compute_stmts_reachability([_Stmt|Stmts]) :-
	compute_stmts_reachability(Stmts).

% Guards
compute_guard_reachability(expGuard(Exp)) :-
	compute_exp_reachability(Exp).
compute_guard_reachability(andGuard(G1,G2)) :-
	compute_guard_reachability(G1),
	compute_guard_reachability(G2).
compute_guard_reachability(claimGuard(_Out)).

% Exps
compute_exps_reachability([]).
compute_exps_reachability([Exp|Exps]) :-
	compute_exp_reachability(Exp),
	compute_exps_reachability(Exps).

compute_exp_reachability(new(CN,Args,_Cog)):- !,
	compute_exps_reachability(Args),
	compute_method_reachability(init,CN).

compute_exp_reachability(newcog(CN,Args,_Cog)) :- !,
	compute_exps_reachability(Args),
	compute_method_reachability(init,CN).

compute_exp_reachability(getExp(Exp)):-!,
	compute_exp_reachability(Exp).

compute_exp_reachability(case(Exp,Branches)) :- !,
	compute_exp_reachability(Exp),
	compute_branches_reachability(Branches).

compute_exp_reachability(apply(Call)) :- !,
	Call =..[Function|Args],
	compute_exps_reachability(Args),
	compute_def_reachability(Function).

compute_exp_reachability(let((_,_),Exp1,Exp2)) :- !,
	compute_exp_reachability(Exp1),
	compute_exp_reachability(Exp2).

compute_exp_reachability(cons(ConsExp)) :- !,
	ConsExp =..[_Cons|ConsExpArgs],
	compute_exps_reachability(ConsExpArgs).

compute_exp_reachability(MethodCall) :-
	MethodCall =..[XSync,Exp,InterfacesList,MN,Args],
	member(XSync,[syncCall,asyncCall]), !,
	compute_exps_reachability(Args),
	compute_exp_reachability(Exp),
	obtainImplementingInterfaces(InterfacesList,Interfaces),
	obtainImplementingClasses(Interfaces,ClassesList),
	compute_virtual_methods_reachability(ClassesList,MN).

compute_exp_reachability(Op) :-
	Op =..[Operator|Exps],
	(is_relational_op(Operator) ; Operator== neg ; is_arithmetic_op(Operator)), !,
	compute_exps_reachability(Exps).
compute_exp_reachability(_).
%branches
compute_branches_reachability([]).
compute_branches_reachability([branch(_Patt,Exp)|Branches]) :-
	compute_exp_reachability(Exp),
	compute_branches_reachability(Branches).

is_method_entry(Entry,MethodName,ClassName) :-
	ut_atom_codes(Entry,EntryCodes),
	ut_split_path_by_code(EntryCodes,46,[MethodName,ClassName]).

%auxiliary operations. 
print_reachable_methods :-
	reachable_method(MethodName,Class),
	format('~w.~w~n',[Class,MethodName]),
	fail.
print_reachable_methods.

print_reachable_defs:-
	reachable_def(Def),
	format('~w~n',[Def]),
	fail.
print_reachable_defs.

% Generalization of ut_split_path
% Should it be included in utils?
ut_split_path_by_code([],_Code,[]).
ut_split_path_by_code(CP,Code,CPL) :-
	ut_append(Pre,[Code|Post],CP), !, 
	atom_codes(Dir,Pre),
	CPL = [Dir|More],
	ut_split_path_by_code(Post,Code,More).
ut_split_path_by_code(CP,_Code,[CPA]) :-
	atom_codes(CPA,CP).
