/** <module> Costabs auxiliary predicates

This module provides a set of auxiliary predicates for Costabs
  
@author Miky Zamalloa
@license GPL
*/

:- module(auxiliary_preds,[is_relational_exp/1, is_relational_op/1, is_arithmetic_op/1,
			   negate_relop/2, expand_normalize_entries/2, remove_outputs/2,
			   obtainImplementingInterfaces/2, obtainImplementingClasses/2,
			   mappingVariablesEntryProgram/2]).

:- use_module(stdlib(utils),[ut_flat_list/2,ut_member/2,ut_sort_rdup/2]).
:- use_module(loaded_abs_db,[interface/3,class/3,method/4,def/4]).
:- use_module(abs2rbr_compiler,[build_method_rule_name/3]).
:- use_module(crs_abs_db,[abs_eq/7]).
:- use_module(rbr_abs_db,[builtin_or_abs_rule/6]).


remove_outputs([],[]).
remove_outputs([T|Ts],[T_p|Ts_p]) :-
	T =..[F,In,_Out],
	T_p =..[F|In],
	remove_outputs(Ts,Ts_p).


expand_normalize_entries(all,ENEs) :- !,
	expand_normalize_entries(all-defs,DefEs),
	expand_normalize_entries(all-methods,MethodEs),
	append(DefEs,MethodEs,ExpandedEs),
	normalize_entries(ExpandedEs,ENEs).
expand_normalize_entries(all-defs,ENEs) :- !,
	findall(FN,def(FN,_,_,_),ExpandedEs),
		normalize_entries(ExpandedEs,ENEs).
expand_normalize_entries(all-methods,ENEs) :- !,
	findall(RN, (method(MSig,CN,_,_),
		     build_method_rule_name(MSig,CN,RN)),
		ExpandedEs),
	normalize_entries([main|ExpandedEs],ENEs).
expand_normalize_entries(Es,NormalEs) :-
	normalize_entries(Es,NormalEs).


%% The entry for PUBS must be a term of the form 'f(A,B)', so if we
%  only provide 'f' we look at its associated cost equation and
%  complete it to 'f(A,B)'.
normalize_entry(E,E) :- compound(E), !.
normalize_entry(E,NE) :-
	abs_eq(E,In,_,_,_,_,_), !,
	length(In,N),
	functor(NE,E,N).

normalize_entries([],[]).
normalize_entries([E|Es],[NE|NEs]) :-
	(normalize_entry(E,NE) -> true ; NE = E),
	normalize_entries(Es,NEs).


is_relational_exp(Exp) :-
	functor(Exp,Op,2),
	member(Op,[eq,neq,gt,geq,lt,leq]).

is_relational_op(Op) :- member(Op,[eq,neq,gt,geq,lt,leq]).

is_arithmetic_op(Op) :- member(Op,[+,-,*,/,mod]).

negate_relop(eq,neq).
negate_relop(neq,eq).
negate_relop(leq,gt).
negate_relop(lt,geq).
negate_relop(geq,lt).
negate_relop(gt,leq).

%predicates that obtain information from the database
obtainImplementingInterfaces(InterfacesIn,InterfacesOut):-
	obtainImplementingInterfaces_(InterfacesIn,[],InterfacesOut).

obtainImplementingInterfaces_([],Accum,InterfacesOut):-
	ut_flat_list(Accum,InterfacesOut).
obtainImplementingInterfaces_([I|IList],Accum,InterfacesOut):-
	findall(SubInterface,(interface(SubInterface,List,_),ut_member(I,List)),SubInterfaces),
	Accum2=[SubInterfaces,[I|Accum]],
	obtainImplementingInterfaces_(IList,Accum2,InterfacesOut).

obtainImplementingClasses(InterfacesIn,ClassesOut):-
	obtainImplementingClasses_(InterfacesIn,[],ClassesOut).

obtainImplementingClasses_([],Accum,ClassesOut):-
	ut_flat_list(Accum,ClassesAux),
	ut_sort_rdup(ClassesAux,ClassesOut).
obtainImplementingClasses_([I|IList],Accum,ClassesOut):-
	findall(Class,(class(Class,List,_),ut_member(I,List)),Classes),
	Accum2=[Classes|Accum],
	obtainImplementingClasses_(IList,Accum2,ClassesOut).

%this predicate relates variables of an entry in the RBR
%with the corresponding parameters in the ABS program

mappingVariablesEntryProgram(Entry,NewEntry):-
	Entry=..[RuleName|Args],
	builtin_or_abs_rule(RuleName,This,In,_,_,_),
	generateArgs(This,In,Args,NewInArgs),
	NewEntry=..[RuleName|NewInArgs].

generateArgs(none,In,Args,NewInArgs):-!,
	generateArgs_aux(In,Args,NewInArgs).
generateArgs(This,In,Args,NewInArgs):-
	This=..[_|[A]],
	Args=[A|RArgs],
	generateArgs_aux(In,RArgs,NArgs),
	NewInArgs=[A|NArgs].

generateArgs_aux([],R,R):-!.
generateArgs_aux([A|As],[B|Bs],[Value|NBs]):-
	A=..[_|[Value]],B=Value,
	generateArgs_aux(As,Bs,NBs).
