
/** <module> Class Invariants

This module provides the support for the generation and loading of the
Costabs class invariants. There are basically two kind of class
invariants: 1) Those that are specified at method entries (the former
"requires" invariants), and, 2) those that are specified at await or
release program points (former "ensures" invariants). They are
represented resp. as entry_invariant/2 and await_invariant/3
facts. During the abs2rbr compilation the source annotations are
loaded as entry_anns/3 and await_anns/3 facts. Then, predicate
load_class_invariants/0 parses each annotation producing the
corresponding entry_invariant/2, await_invariant/3 and inv_param/2
facts.

@author Miky Zamalloa
@license GPL
*/

:- module(class_invariants,[add_entry_anns/3, add_await_anns/3, add_inv_param/2,
			    add_entry_invariant/2, add_await_invariant/3,
			    entry_invariant/2, await_invariant/3, class_inv_params/2,
			    load_class_invariants/0, cleanup_class_invs/0, print_class_invs/0]).

:- use_module(loaded_abs_db,[class/3]).
:- use_module(stdlib(profiling),[profiling_start_timer/1,profiling_stop_timer/2]).
:- use_module(costabs_output_manager,[info_message/2]).


%%%%%%%%%%%%%%%% Class invariants DB %%%%%%%%%%%%%%%%%%%

:- dynamic entry_anns/3. % entry_anns(CN,RN,Anns)
:- dynamic await_anns/3. % await_anns(CN,AwaitId,Anns)

:- dynamic inv_param/2. % inv_param(CN,Param) -> Param is a parameter in class invariant of CN
:- dynamic inv_params/2. % inv_params(CN,Params) -> Params is the list of inv params of class CN
:- dynamic entry_invariant/2. % entry_invariant(RN,Inv)
:- dynamic await_invariant/3. % await_invariant(AwaitId,UnmodifiedFields,Inv)


add_entry_anns(CN,RN,Anns) :- assertz(entry_anns(CN,RN,Anns)).

add_await_anns(CN,AwId,Anns) :- assertz(await_anns(CN,AwId,Anns)).

add_inv_param(CN,PN) :- inv_param(CN,PN), !. % We avoid repetitions
add_inv_param(CN,PN) :- assertz(inv_param(CN,PN)).

add_entry_invariant(RN,Inv) :- assertz(entry_invariant(RN,Inv)).

add_await_invariant(AwId,UnmFs,Inv) :- assertz(await_invariant(AwId,UnmFs,Inv)).

class_inv_params(CN,Params) :- inv_params(CN,Params), !.
class_inv_params(CN,Params) :-
	findall(PN,inv_param(CN,PN),Params),
	assertz(inv_params(CN,Params)).

cleanup_class_invs :-
	retractall(entry_anns(_,_,_)),
	retractall(await_anns(_,_,_)),
	retractall(inv_param(_,_)),
	retractall(inv_params(_,_)),
	retractall(entry_invariant(_,_)),
	retractall(await_invariant(_,_,_)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%% Loading and parsing ABS source annotations %%%%%%%%%

load_class_invariants :- 
	profiling_start_timer(c_invs),
	% Here we call Antonio's analysis
	% class_invariants_analysis(...),
	entry_anns(CN,RN,Anns),
	parse_entry_anns(CN,RN,Anns),
	fail.
load_class_invariants :-
	await_anns(CN,AwId,Anns),
	parse_await_anns(CN,AwId,Anns),
	fail.
load_class_invariants :-
	profiling_stop_timer(c_invs,T),
	info_message(c_invs,[T]).


%% This parses the list of annotations Anns of rule RN entry of class
%  CN. The resulting entry invariant is asserted.
parse_entry_anns(CN,RN,Anns) :-
	parse_entry_anns_(CN,Anns,Inv),
	(Inv \== [] -> assertz(entry_invariant(RN,Inv))
	             ; true).

parse_entry_anns_(CN,[Ann|Anns],[InvCond|Inv]) :-
	parse_ann(CN,Ann,InvCond), 
	parse_entry_anns_(CN,Anns,Inv).
parse_entry_anns_(_,[],[]).


%% This parses the list of annotations Anns in await Id of class
%  CN. The resulting await invariant is asserted.
parse_await_anns(CN,Id,Anns) :-
	parse_await_anns_(CN,Anns,UnmodFs,Inv),
	((UnmodFs \== []; Inv \== []) -> assertz(await_invariant(Id,UnmodFs,Inv))
	                               ; true).

parse_await_anns_(CN,[Ann|Anns],[UnmField|UnmFs],Inv) :-
	parse_unm_field_ann(CN,Ann,UnmField), !,
	parse_await_anns_(CN,Anns,UnmFs,Inv).
parse_await_anns_(CN,[Ann|Anns],UnmFs,[InvCond|Inv]) :-
	parse_ann(CN,Ann,InvCond), 
	parse_await_anns_(CN,Anns,UnmFs,Inv).
parse_await_anns_(_,[],[],[]).

% Particular case of old(f) == f annotation
% Check/update when Antonio fixes the Prolog backend
parse_unm_field_ann(_CN,Ann,field(FName)) :-
	Ann = eq(apply(old(l(FName))),l(FName)).


%%%%%%%%%%%%%%% Parsing single annotations %%%%%%%%%%%%%%%

parse_ann(CN,Ann,InvCond) :-
	handle_ann_vars(CN,Ann,Ann_p),
	relop_to_constraint(Ann_p,InvCond) ,!.
parse_ann(_,Ann,_) :- throw(costabs_exception(not_supported_annotation(Ann))).

handle_ann_vars(CN,Ann,Ann_p) :-
	Ann =..[Op,V1,V2],
	handle_ann_var(CN,V1,W1),
	handle_ann_var(CN,V2,W2),
	Ann_p =..[Op,W1,W2].

handle_ann_var(CN,l(FN),field(FN)) :- % Remove this when Antonio fixes the Prolog backend
	class(CN,_,Fields),
	member(field(FN,_,_),Fields), !.
handle_ann_var(CN,apply(max(l(FN))),max(FN)) :- !, add_inv_param(CN,max(FN)). % Remove this when Antonio fixes the Prolog backend
handle_ann_var(CN,apply(max(field(FN))),max(FN)) :- !, add_inv_param(CN,max(FN)).
handle_ann_var(CN,apply(min(l(FN))),min(FN)) :- !, add_inv_param(CN,min(FN)). % Remove this when Antonio fixes the Prolog backend
handle_ann_var(CN,apply(min(field(FN))),min(FN)) :- !, add_inv_param(CN,min(FN)).
handle_ann_var(_,V,V). % V is actually not always a var but it can also be a constant

relop_to_constraint(eq(X,Y),'='(X,Y)).
relop_to_constraint(lt(X,Y),'>='(Y,X+1)).
relop_to_constraint(leq(X,Y),'>='(Y,X)).
relop_to_constraint(gt(X,Y),'>='(X,Y+1)).
relop_to_constraint(geq(X,Y),'>='(X,Y)).


%%%%%%%%%%%%%%%% Auxiliar predicates %%%%%%%%%%%%%%%

print_class_invs :-
	ho_print_pred(inv_param/2),
	ho_print_pred(entry_invariant/2),
	ho_print_pred(await_invariant/3).

ho_print_pred(F/N) :-
	functor(Call,F,N),
	catch(call(Call),error(_,_),fail), % This is to avoid errors when no clause is found
	write(Call),nl,
	fail.
ho_print_pred(_).

	
