/** <module> Cost models of Costabs

This module implements the different Costabs cost models. Essentially,
predicate apply_cost_model/2, given a list of RBR instructions,
returns in the second argument a (possibly symbolic) cost expression
according to the specified cost model (param 'cost_model'), which will
be directly put on the corresponding cost equation.
  
@author Miky Zamalloa
@license GPL
*/

:- module(cost_models,[apply_cost_model/2]).

:- use_module(costabs_params,[costabs_param/2]).
:- use_module(rbr_abs_db,[abs_rule/6]).
		     

apply_cost_model(Trace,CostExp) :-
	costabs_param(cost_model,Model),
	apply_cost_model(Model,Trace,CostExp).


apply_cost_model(steps,Trace,CostExp) :- length(Trace,CostExp).
apply_cost_model(memory,Trace,CostExp) :- memory_cost_model(Trace,CostExp,0).
apply_cost_model(objects,Trace,CostExp) :- objects_cost_model(Trace,CostExp,0).
apply_cost_model(task_level,Trace,CostExp) :- tasklevel_cost_model(Trace,CostExp,0).
apply_cost_model(cogs,Trace,CostExp) :- cogs_cost_model(Trace,CostExp,0).
apply_cost_model(termination,_,1).


memory_cost_model([],Ac,Ac).
memory_cost_model([cons(_,T)|RTrace],CostExp,Ac) :- !,
	functor(T,Co,_),
	(costabs_param(symb_or_num,symb) -> X = c(size(Co))
	                                  ; X = 1),
	Ac_p = Ac + X,
	memory_cost_model(RTrace,CostExp,Ac_p).
memory_cost_model([_|RTrace],CostExp,Ac) :-
	memory_cost_model(RTrace,CostExp,Ac).


objects_cost_model([],Ac,Ac).
objects_cost_model([new(CN)|RTrace],CostExp,Ac) :- !,
	(costabs_param(symb_or_num,symb) -> X = c(object(CN))
	                                  ; X = 1),
	Ac_p = Ac + X,
	objects_cost_model(RTrace,CostExp,Ac_p).
objects_cost_model([newcog(CN)|RTrace],CostExp,Ac) :- !,
	(costabs_param(symb_or_num,symb) -> X = c(object(CN))
	                                  ; X = 1),
	Ac_p = Ac + X,
	objects_cost_model(RTrace,CostExp,Ac_p).
	
objects_cost_model([_|RTrace],CostExp,Ac) :-
objects_cost_model(RTrace,CostExp,Ac).

cogs_cost_model([],Ac,Ac).
cogs_cost_model([newcog(_CN)|RTrace],CostExp,Ac) :- !,
	Ac_p = Ac + 1,
	cogs_cost_model(RTrace,CostExp,Ac_p).
cogs_cost_model([_|RTrace],CostExp,Ac) :-
	cogs_cost_model(RTrace,CostExp,Ac).


tasklevel_cost_model([],Ac,Ac).
tasklevel_cost_model([call(RN)|RTrace],CostExp,Ac) :-
	abs_rule(RN,_,_,_,_,[method,_,_]), !,
	Ac_p = Ac + 1,
	tasklevel_cost_model(RTrace,CostExp,Ac_p).
tasklevel_cost_model([_|RTrace],CostExp,Ac) :-
	tasklevel_cost_model(RTrace,CostExp,Ac).
