/** <module> The COSTABS driver

This module implements the driver of COSTABS. The process basically
consists of:

  1) Load the ABS input program by means of
  abs_program_loader:load_abs_program/1. This first calls the
  generateProlog script, and then reads the generated file and asserts
  its contents in loaded_abs_db. There is also the possiblity of not
  calling generateProlog and just assume that the current ABS program
  is loaded in the abs.pl file.

  2) Compile the ABS into the RBR(abs) by means of
  abs2rbr_compiler:compile_to_rbr/0. The resulting RBR(abs) is stored
  in rbr_abs_db.

  3) Generate the CRS from the RBR(abs) by means of predicate
  rbr2crs:rbr2crs/0. The resulting CRS is stored in crs_abs_db.

  4) Once the CRS is generated PUBS will solve it. Module crs_solver
  takes care of communicating with PUBS by means of predicate
  solve_crs/1. The obtained UBs are stored in module ubs_db

  5) Finally the results are reported by costabs_output_manager

There are three drivers with different combination of args:
  - abs2rbr: It just performs steps 1) and 2)
  - abs2crs: It performs steps 1), 2) and 3)
  - costabs: It performs all steps
  
Example call:
  ?- costabs(['MyPeerToPeer.abs'],['Node.reqFile'],[debug = yes]).
  
@author Miky Zamalloa
@license GPL

*/

:- module(costabs_driver,[costabs/0, costabs/1, costabs/2, costabs/3,
			  abs2rbr/0, abs2rbr/1, abs2rbr/2, abs2rbr/3,
			  abs2crs/0, abs2crs/1, abs2crs/2, abs2crs/3]).


:- use_module(abs_program_loader,[load_abs_program/2]).
:- use_module(loaded_abs_db,[cleanup_ladb/0]).
:- use_module(abs2rbr_compiler,[compile_to_rbr/0]).
:- use_module(rbr_abs_db,[cleanup_rbrdb/0, abs_rule/6]).
:- use_module(rbr2crs,[rbr2crs/0, cleanup_rbr2crs/0]).
:- use_module(crs_abs_db,[cleanup_crsdb/0]).
:- use_module(costabs_params,[set_default_costabs_params/0, 
                              set_costabs_params/1, costabs_param/2,
			      cleanup_costabs_params/0]).
:- use_module(costabs_output_manager,[info_message/2,write_file/2,write_ubs_to_xml_file/1]).
:- use_module(class_invariants,[load_class_invariants/0, cleanup_class_invs/0]).
:- use_module(ubs_db,[costabs_ub/3,cleanup_ubs/0]).
:- use_module(auxiliary_preds,[expand_normalize_entries/2]).
:- use_module(crs_solver,[solve_crs/1]).

:- use_module(stdlib(profiling),[profiling_initialize/0,profiling_start_timer/1,
				 profiling_stop_timer/2]).



costabs :- costabs([],all-defs,[]).
costabs(Entries) :- costabs([],Entries,[]).
costabs(AbsFNs,Entries) :- costabs(AbsFNs,Entries,[]).
costabs(AbsFNs,Entries,Params) :-
	catch(costabs_(AbsFNs,Entries,Params),  
	      costabs_exception(E),info_message(costabs_exception,[E])).
	%cleanup_costabs.

costabs_(AbsFNs,Entries,Params) :-

	% Steps 1, 2 and 3 above
	abs2crs(AbsFNs,Entries,Params),

	% Step 4: Solving the CRS in PUBS
	profiling_start_timer(pubs),
	solve_crs(Entries), 
	profiling_stop_timer(pubs,T_pubs),
	info_message(pubs,[T_pubs]),

	% Step 5: Reporting results
	report_ubs(Entries),

	% Step 6: Export results to xml
	write_ubs_to_xml_file(Entries).


abs2crs :- abs2crs([],all-defs,[]).
abs2crs(Entries) :- abs2crs([],Entries,[]).
abs2crs(AbsFNs,Entries) :- abs2crs(AbsFNs,Entries,[]).
abs2crs(AbsFNs,Entries,Params) :-
	% Steps 1 and 2 above
	abs2rbr(AbsFNs,Entries,Params),

	% Loading class invariants
	load_class_invariants,
	
	% Generating the CRS
	rbr2crs,
	write_file(crs,[]).


abs2rbr :- abs2rbr([],all-defs,[]).
abs2rbr(Entries) :- abs2rbr([],Entries,[]).
abs2rbr(AbsFNs,Entries) :- abs2rbr(AbsFNs,Entries,[]).
abs2rbr(AbsFNs,Entries,Params) :-
	% Initializing
	init_costabs, 
	set_costabs_params(Params),

	% Loading the ABS program
	profiling_start_timer(load_abs),
	load_abs_program(AbsFNs,Entries),
	profiling_stop_timer(load_abs,T_load),
	info_message(load_abs,[T_load]),

	% Compiling to the RBR
	profiling_start_timer(rbr),
	compile_to_rbr,
	profiling_stop_timer(rbr,T_rbr),
	info_message(gen_rbr,[T_rbr]),
	check_entries_exist(Entries),
	write_file(rbr,[]).


check_entries_exist(all) :- !.
check_entries_exist(all-defs) :- !.
check_entries_exist(all-methods) :- !.
check_entries_exist([]) :- !.
check_entries_exist([Entry|Entries]) :-
	abs_rule(Entry,_,_,_,_,_), !,
	check_entries_exist(Entries).
check_entries_exist([Entry|_Entries]) :-
	throw(costabs_exception(inexistent_entry(Entry))).


report_ubs(Entries) :-
	expand_normalize_entries(Entries,NormalEntries),
	member(Entry,NormalEntries),
	(costabs_param(cost_model,termination) -> (
		functor(Entry,F,N),
		costabs_ub(Entry,_,YesOrNo),
	 	info_message(termin,[F/N,YesOrNo]))
	; 
	        info_message(ub,[Entry])),
	fail.
report_ubs(_).

cleanup_costabs :-
	cleanup_ladb,
	cleanup_rbrdb,
	cleanup_crsdb,
	cleanup_costabs_params,
	cleanup_class_invs,
	cleanup_rbr2crs,
	cleanup_ubs.

init_costabs :-
	cleanup_costabs,
	profiling_initialize,
	set_default_costabs_params.
