/** <module> The COSTABS output manager

This module manages the output of costabs.  It has two main
predicates: info_message(Message_id,Arguments) Prints the message
corresponding to the given message id if and only if the program is
being executed with the needed verbosity.

write_file(File_id,Arguments) Writes the file corresponding to the
given file id if and only if the program is being executed with the
needed verbosity.

All the output is done through this module. Each message has an
info_message clause. We can consider that each message needs a set of
requisites to be shown.  Those requisites are specified in the
predicate check_requisites/1.  That way each message can have its
custom display requisites or share requisites with other messages of
the similar characteristics.

@author Antonio Flores and Miky Zamalloa
@license GPL

*/

:- module(costabs_output_manager,[info_message/2,write_file/2,write_ubs_to_xml_file/1]).


:- use_module(costabs_params,[costabs_param/2]).
:- use_module(rbr_abs_db,[print_rbrdb/0,abs_rule/6]).
:- use_module(crs_abs_db,[print_crsdb/0,print_crs_pubs/0]).
:- use_module(ubs_db,[costabs_ub/3]).
:- use_module(auxiliary_preds,[expand_normalize_entries/2,mappingVariablesEntryProgram/2]).

:- use_module(stdlib(utils),[ut_atom_concat/3]).



% Constants
default_output_path('/tmp/costabs/').


% Interface
info_message(welcome,[]) :- 
	check_requisites(welcome), !,
	format("This is COSTABS ~n",[]).
info_message(load_abs,[T]) :- 
	check_requisites(load_abs), !,
	format("~nAbs program loaded in ~0f ms.~n",[T]).
info_message(gen_rbr,[T]) :- 
	check_requisites(rbr), !,
	format("~nRule based representation generated in ~0f ms.~n",[T]).
info_message(abst_comp,[T]) :- 
	check_requisites(abst_comp), !,
	format("~nAbstract compilation performed in ~0f ms.~n",[T]).
info_message(sabu,[T]) :- 
	check_requisites(sabu), !,
	format("~nSize analysis performed in ~0f ms.~n",[T]).
info_message(f_assers,[T]) :- 
	check_requisites(f_assers), !,
	format("~nField assertions analysis performed in ~0f ms.~n",[T]).
info_message(c_invs,[T]) :- 
	check_requisites(c_invs), !,
	format("~nClass invariants generated and loaded in ~0f ms.~n",[T]).
info_message(pubs,[T]) :- 
	check_requisites(pubs), !,
	format("~nCost relation system solved by PUBS in ~0f ms.~n",[T]).
info_message(ub,[Entry]) :-
	costabs_param(cost_centers,no),
	check_requisites(ub), !,
	costabs_ub(Entry,_,UB),
	format('~nUB for ',[]),
	mappingVariablesEntryProgram(Entry,NEntry),
	numbervars([NEntry],0,_),
	numbervars([NEntry,UB],0,_),
	write_term(NEntry,[numbervars(true),quoted(true)]), 
	format(' = ',[]),
	write_term(UB,[numbervars(true),quoted(true)]), nl.

info_message(ub,[Entry]) :-
	costabs_param(cost_centers,yes),
	check_requisites(ub), !,
	%numbervars([Entry],0,_),
	format('~nUB for ',[]),
        mappingVariablesEntryProgram(Entry,NEntry),
	numbervars([NEntry],0,_),
	write_term(NEntry,[numbervars(true),quoted(true)]),
	functor(NEntry,RN,_),
	((abs_rule(RN,_,_,_,_,[_,none,_]),RN \== main) -> (
	 	format(' = ',[]),
		costabs_ub(NEntry,_,UB),
		numbervars([NEntry,UB],0,_),
	        write_term(UB,[numbervars(true),quoted(true)]), nl)
	;(
	        format(' per cost-center = ~n',[]),
	        write_ub_per_cc(NEntry), nl)).
info_message(termin,[Entry,YesOrNo]) :- 
	check_requisites(termin), !,
	(YesOrNo = yes ->
	     format('~n~q is terminating~n',[Entry])
	;
	     format('~n~q is non-terminating~n',[Entry])).


info_message(not_supported_term,[Term]) :- 
	check_requisites(warning), !,
	format(user_error,'~nTerm not supported in abs.pl ~q~n',[Term]).
info_message(not_supported_size_abstraction,[B]) :- 
	check_requisites(warning), !,
	format(user_error,"Size abstraction not supported for ~q~n",[B]).

info_message(found_variable_error,[]) :- 
	check_requisites(warning), !,
	format(user_error,"Found a variable during varnames_to_prologvars!",[]).
info_message(field_vars_info,[FieldVars]) :- 
	check_requisites(debug_info), !,
	writeln(FieldVars).

% Exceptions messages
info_message(costabs_exception,[wrong_costabs_parameter(Name,Value)]) :- 
	check_requisites(exception), !,
	format(user_error,"Exception: Illegal value of parameter ~w: '~w'~n",[Name,Value]).

info_message(costabs_exception,[abs_pl_file_not_found(Path)]) :- 
	check_requisites(exception), !,
	format(user_error,"Exception: The intermediate file '~w' was not found~n",[Path]).

info_message(costabs_exception,[wrong_abs_file(AbsPath)]) :- 
	check_requisites(exception), !,
	format(user_error,"Exception: The Abs file '~w' does not exist~n",[AbsPath]).

info_message(costabs_exception,[inexistent_entry(Entry)]) :- 
	check_requisites(exception), !,
	format(user_error,"Exception: The specified entry '~w' does not exist~n",[Entry]).
info_message(costabs_exception,[not_supported_stmt(Stmt,RName)]) :- 
	check_requisites(exception), !,
	format("~nException: Not supported statement '~w' found during the RBR compilation
	        of rule '~w'~n",[Stmt,RName]).

% In case an exception message has not been specified it will still be shown
info_message(costabs_exception,[E]) :- 
	check_requisites(exception), !,
	format(user_error,"Exception: ~w~n",[E]).

info_message(_,_).


write_ub_per_cc(Entry) :-
	costabs_ub(Entry,CC,UB),
	nl,write_term(UB,[numbervars(true),quoted(true)]),
	format(" within cost-center ~q~n",[CC]),
	fail.
write_ub_per_cc(_).

write_file(rbr,[]) :-
	check_requisites(debug_info), !,
	default_output_path(DefaultPath),
	ut_atom_concat(DefaultPath,'rbr.rbr',FileName),
	open(FileName,write,SOut),
	set_output(SOut),
	print_rbrdb,
	set_output(user_output),
	close(SOut),
	(check_requisites(rbr_to_file) -> 
	        format("~nRule based representation stored in ~w~n",[FileName])
	;
	        true).
write_file(crs,[]) :-
	check_requisites(debug_info), !,
	default_output_path(DefaultPath),
	ut_atom_concat(DefaultPath,'crs.crs',FileName),
	open(FileName,write,SOut),
	set_output(SOut),
	(costabs_param(crs_format,pretty) -> print_crsdb
	                                   ; print_crs_pubs),
	set_output(user_output),
	close(SOut),
	(check_requisites(crs_to_file) ->
	        format("~nCost relation system stored in ~w~n",[FileName])
	;
	        true).
write_file(_,_).


% Configuration of what it is written according to the verbosity flag

% Always
check_requisites(welcome).
check_requisites(exception).

% Verbosity 1 or greater
check_requisites(load_abs) :-
	costabs_param(verbosity,N),
	N >= 1.
check_requisites(rbr) :-
	costabs_param(verbosity,N),
	N >= 1.
check_requisites(abst_comp) :-
	costabs_param(verbosity,N),
	N >= 1.
check_requisites(sabu) :-
	costabs_param(verbosity,N),
	N >= 1.
check_requisites(f_assers) :-
	costabs_param(verbosity,N),
	N >= 1.
check_requisites(c_invs) :-
	costabs_param(verbosity,N),
	N >= 1.
check_requisites(pubs) :-
	costabs_param(verbosity,N),
	N >= 1.
check_requisites(ub) :-
	costabs_param(verbosity,N),
	N >= 0.
check_requisites(termin) :-
	costabs_param(verbosity,N),
	N >= 1.


% Verbosity 2 or greater
check_requisites(warning) :- check_requisites(debug_info).

check_requisites(rbr_to_file) :-
	costabs_param(verbosity,N),
	N >= 2.
check_requisites(crs_to_file) :-
	costabs_param(verbosity,N),
	N >= 2.

% Debug
check_requisites(debug_info) :-
	costabs_param(debug,yes).

write_ubs_to_xml_file(Entries):-
	open('/tmp/costabs/abs.xml',write,SOut),
	format(SOut,"<?xml version=\"1.0\"?>~n",[]),
	format(SOut,"<costabs>~n",[]),
	write_ubs_to_xml_file_aux(SOut,Entries),
	format(SOut,"</costabs>~n",[]),
	close(SOut).

% Exporting upper bounds results to xml file
write_ubs_to_xml_file_aux(SOut,Entries) :-
	expand_normalize_entries(Entries,NormalEntries),
	member(Entry,NormalEntries),
	(costabs_param(cost_model,termination) -> (
		functor(Entry,F,N),
		costabs_ub(Entry,_,YesOrNo),
	 	write_termination_to_xml(SOut,[F/N,YesOrNo]))
	; 
	        write_ub_to_xml(SOut,[Entry])),
	fail.
write_ubs_to_xml_file_aux(_,_).

write_ub_to_xml(SOut,[Entry]) :-
	costabs_param(cost_centers,no),
	costabs_ub(Entry,_,UB),
	numbervars([Entry,UB],0,_),
	format(SOut,"<elem>~n",[]),
	format(SOut,"<name>",[]),
	write_term(SOut,Entry,[numbervars(true),quoted(true)]), 
	format(SOut,"</name>~n",[]),
	format(SOut,"<ub>",[]),
	write_term(SOut,UB,[numbervars(true),quoted(true)]), 
	format(SOut,"</ub>~n",[]),
	format(SOut,"</elem>~n",[]).

write_ub_to_xml(SOut,[Entry]) :-
	costabs_param(cost_centers,yes),
	numbervars([Entry],0,_),
	format(SOut,"<elem>~n",[]),
	format(SOut,"<name>",[]),
	write_term(SOut,Entry,[numbervars(true),quoted(true)]),
	format(SOut,"</name>~n",[]),
	format(SOut,"<ub>",[]),
	functor(Entry,RN,_),
	((abs_rule(RN,_,_,_,_,[_,none,_]),RN \== main) -> ( % RN is a function
		costabs_ub(Entry,_,UB),
	        write_term(SOut,UB,[numbervars(true),quoted(true)]))
	;(
	        write_ub_per_cc_xml(SOut,Entry))),
	format(SOut,"</ub>~n",[]),
	format(SOut,"</elem>~n",[]).

write_termination_to_xml(SOut,[Entry,YesOrNo]) :- 
	format(SOut,"<elem>~n",[]),
	format(SOut,"<name>",[]),
	write_term(SOut,Entry,[numbervars(true),quoted(true)]),
	format(SOut,"</name>~n",[]),
	format(SOut,"<termin>",[]),
	(YesOrNo = yes ->
	     format(SOut,'yes',[])
	;
	     format(SOut,'no',[])),
	format(SOut,"</termin>~n",[]),
	format(SOut,"</elem>~n",[]).

write_ub_per_cc_xml(SOut,Entry) :-
	costabs_ub(Entry,CC,UB),
	write_term(SOut,UB,[numbervars(true),quoted(true)]),
	format(SOut," within cost-center ~q~n",[CC]),
	fail.
write_ub_per_cc_xml(_,_).

