/** <module> CRS(abs) Database

This module provides an interface for the DB in which the current
CRS(abs) for the current ABS program is stored.
  
@author Miky Zamalloa
@license GPL
*/


:- module(crs_abs_db,[abs_eq/7, builtin_or_abs_eq/7, add_abs_eq/7, add_rels_to_abs_eqs/3,
		      print_crsdb/0, cleanup_crsdb/0, print_crs_pubs/0]).

:- use_module(rbr_abs_db,[call_rel/2]).
:- use_module(auxiliary_preds,[remove_outputs/2]).


%% The RBR(abs) is represented as a set of facts of the form
%          abs_eq(Name,In,Out,Exp,Calls,Constraints,Info)
:- dynamic abs_eq/7.

add_abs_eq(A,B,C,D,E,F,G) :- assertz(abs_eq(A,B,C,D,E,F,G)).


%% Add the relations Rels to all rules that call F (only when it is
%  not the last call in the body)
add_rels_to_abs_eqs(F,Args,Rels) :-
	CallTemplate =..[F,_In,_Out],
	call_rel(Caller,F),
	% We have to do a findall because new abs_eq facts are added in the process
	findall(abs_eq(Caller,CallIn,CallOut,Exp,Calls,Phi,[EqId]),
		abs_eq(Caller,CallIn,CallOut,Exp,Calls,Phi,[EqId]),
		AbsEqs),
	member(Eq,AbsEqs),
	Eq = abs_eq(Caller,CallIn,CallOut,Exp,Calls,Phi,[EqId]),
	% There can be more than one call to F -> We need the rels for all of them
	% It fails if the only call to F is the last call in Calls
	rels_forall_calls(CallTemplate,Calls,Args,Rels,AllRels),
	append(AllRels,Phi,Phi_p),	
%	member_but_not_last(CallTemplate,Calls),
%	append(In,Out,Args), % Here the vars are unified
%	append(Rels,Phi,Phi_p),
	Eq_p = abs_eq(Caller,CallIn,CallOut,Exp,Calls,Phi_p,[EqId]),
	replace_abs_eq(Caller,EqId,Eq_p),
	fail.
add_rels_to_abs_eqs(_,_,_).

%% As there can be many calls to F, we have to traverse the calls,
%  and, for each call to F, we must unify the call vars with Args and
%  add the Rels to the result. The last call is not considered and if
%  there is only one call to F it fails.
rels_forall_calls(Call,[Call,Next|Rest],Args,Rels,AllRels) :- !,
	Call =..[_,In,Out],
	copy_term([Args,Rels,Call],[Args_p,Rels_p,Call_p]),
	%[Args,Rels,Call] = [Args_p,Rels_p,Call_p],
	append(In,Out,Args_p),
	(Rest = [] -> AllRels = Rels_p
	            ;(rels_forall_calls(Call_p,[Next|Rest],Args,Rels,RestRels),
		      append(Rels_p,RestRels,AllRels))).
rels_forall_calls(Call,[_,Next|Rest],Args,Rels,AllRels) :-
	(Rest = [] -> AllRels = []
	            ; rels_forall_calls(Call,[Next|Rest],Args,Rels,AllRels)).


replace_abs_eq(Caller,EqId,Eq_p) :-
	retract(abs_eq(Caller,_,_,_,_,_,[EqId])), !,
	assertz(Eq_p).

% Builtin-equations for relational operators
builtin_eq(eq,[X,Y],[Z],0,[],[X = Y, Z = 1]).
builtin_eq(eq,[X,Y],[Z],0,[],[X >= Y+1, Z = 0]).
builtin_eq(eq,[X,Y],[Z],0,[],[Y >= X+1, Z = 0]).
builtin_eq(neq,[X,Y],[Z],0,[],[X = Y, Z = 0]).
builtin_eq(neq,[X,Y],[Z],0,[],[X >= Y+1, Z = 1]).
builtin_eq(neq,[X,Y],[Z],0,[],[Y >= X+1, Z = 1]).
builtin_eq(lt,[X,Y],[Z],0,[],[Y-X >= 1, Z = 1]).
builtin_eq(lt,[X,Y],[Z],0,[],[Y =< X, Z = 0]).
builtin_eq(leq,[X,Y],[Z],0,[],[X =< Y, Z = 1]).
builtin_eq(leq,[X,Y],[Z],0,[],[X-Y >= 1, Z = 0]).
builtin_eq(gt,[X,Y],[Z],0,[],[X-Y >= 1, Z = 1]).
builtin_eq(gt,[X,Y],[Z],0,[],[X =< Y, Z = 0]).
builtin_eq(geq,[X,Y],[Z],0,[],[Y =< X, Z = 1]).
builtin_eq(geq,[X,Y],[Z],0,[],[Y-X >= 1, Z = 0]).
builtin_eq(neg,[X],[Z],0,[],[X = 0, Z = 1]).
builtin_eq(neg,[X],[Z],0,[],[X = 1, Z = 0]).


builtin_or_abs_eq(A,B,C,D,E,F,-1) :-
	builtin_eq(A,B,C,D,E,F).
builtin_or_abs_eq(A,B,C,D,E,F,G) :- abs_eq(A,B,C,D,E,F,G).


cleanup_crsdb :- retractall(abs_eq(_,_,_,_,_,_,_)).

print_crsdb :-
	format('~n~n-----------------------~n',[]),
	format('        CRS(abs)       ~n',[]),
	format('-----------------------~n',[]),
	builtin_or_abs_eq(Name,In,Out,E,Calls,Cs,_),
	numbervars([In,Out,E,Calls,Cs],0,_),
	H =..[Name,In,Out],
	nl, write_term(H,[numbervars(true),quoted(true)]),
	format(' = ',[]),
	write_term(E,[numbervars(true),quoted(true)]),
	findall(C,(member(C,Calls),
		   format(' + ',[]),
		   write_term(C,[numbervars(true),quoted(true)])),
		_),
	format('~n    ~q',[Cs]),
	fail.
print_crsdb.


print_crs_pubs :-
	nl,
	builtin_or_abs_eq(Name,In,Out,E,Calls,Cs,_),
	remove_outputs(Calls,Calls_p),
	numbervars([In,Out,E,Calls_p,Cs],0,_),
	H =..[Name|In],
	write_term(eq(H,E,Calls_p,Cs),[numbervars(true),quoted(true)]),
	format('.~n',[]),
	fail.
print_crs_pubs.


remove_last([],[]).
remove_last([_],[]) :- !.
remove_last([X|Xs],[X|Zs]) :- remove_last(Xs,Zs).

member_but_not_last(X,[X,_|_]).
member_but_not_last(X,[_,Y|R]) :- member_but_not_last(X,[Y|R]).
