/** <module> The Costabs CRS Solver

This module takes care of solving the Costabs CRS. It basically has
two functionalities: 1) It provides a suitable interface of PUBS for
Costabs, thus taking care of all the details behind the way PUBS must
be called within Costabs. 2) When the cost centers functionality is
enabled, it takes care of calling PUBS as many times as existing cost
centers. The obtained UBs (per cost center or not) are stored in
module ubs_db. The Costabs driver is hence responsible of reporting
them.

@author Miky Zamalloa
@license GPL

*/

:- module(crs_solver,[solve_crs/1]).

:- use_module(rbr_abs_db,[abs_rule/6]).
:- use_module(costabs_params,[costabs_param/2]).
:- use_module(auxiliary_preds,[expand_normalize_entries/2,remove_outputs/2]).
:- use_module(loaded_abs_db,[class/3]).
:- use_module(crs_abs_db,[builtin_or_abs_eq/7]).
:- use_module(ubs_db,[store_ub/3]).

:- use_module(costa(parameters_db),[paramdb_set_defaults/0,paramdb_set_list/1]).
:- use_module(stdlib(debugging),[set_verbosity_level/1]).
:- use_module(pubs(pubs_main),[pubs_cr_compute_ub/3,pubs_cr_compute_ub_multiple_entries/3]).
:- use_module(stdlib(debugging),[set_verbosity_level/1]).
:- use_module(pubs(pubs_db),[eval_upper_bound/2,crs_upper_bound/3,init_pubs_db/0,
			     global_non_terminating/1]).
:- use_module(stdlib(ub_writing),[ub_simplify/2]).
:- use_module(pubs(pubs_simplification),[cexpr_to_exp/2]).


solve_crs(Entries) :-
	costabs_param(cost_centers,no), !,
	solve_crs_no_ccs(Entries).
solve_crs(Entries) :- % No sense to use cost centers with termination
	costabs_param(cost_model,termination), !,
	solve_crs_no_ccs(Entries).
solve_crs(Entries) :- 
	solve_crs_ccs(Entries).


solve_crs_ccs(Entries) :-
	expand_normalize_entries(Entries,NormalEntries),
	split_functs_methods(NormalEntries,NormalFEs,NormalMEs),
	solve_crs_no_ccs(NormalFEs),
	crs_to_list_for_pubs(Eqs),
	findall(CC,class(CC,_,_),CCs),
	member(CC,[main|CCs]), % We consider as cost centers all classes + the main
	set_pubs,
	process_eqs_for_cc(CC,Eqs,Eqs_p),
	pubs_cr_compute_ub_multiple_entries(Eqs_p,[],NormalMEs),
	member(Entry,NormalMEs),
	costabs_result_for_entry(Entry,CC),
	fail.
solve_crs_ccs(_).


solve_crs_no_ccs(Entries) :-
	expand_normalize_entries(Entries,NormalEntries),
	set_pubs, 
	crs_to_list_for_pubs(Eqs),
	pubs_cr_compute_ub_multiple_entries(Eqs,[],NormalEntries),
	member(Entry,NormalEntries),
	costabs_result_for_entry(Entry,_),
	fail.
solve_crs_no_ccs(_).

costabs_result_for_entry(Entry,CC) :-
	costabs_param(cost_model,Model),
	member(Model,[steps,memory,objects,task_level,cogs]), !,
	eval_upper_bound(Entry,UB),
	cexpr_to_exp(UB,UB_p),
	ub_simplify(UB_p,UB_pp),
	store_ub(Entry,CC,UB_pp).

costabs_result_for_entry(Entry,_) :-
	costabs_param(cost_model,termination), !,
	functor(Entry,F,N),
	(global_non_terminating(F/N) -> YesOrNo = no ; YesOrNo = yes),
	store_ub(Entry,_,YesOrNo).


set_pubs :-
	set_verbosity_level('0'),
	paramdb_set_defaults,
	init_pubs_db,
	(costabs_param(cost_model,termination) ->
	        paramdb_set_list([check_zero_cost = no,
				  only_rf_scheme = yes,
				  node_count_scheme = no])
	;
	        true).


split_functs_methods([],[],[]).
split_functs_methods([E|Es],[E|FEs],MEs) :-
	functor(E,RN,_),
	abs_rule(RN,_,_,_,_,[_,none,_]),
	RN \== main, !, % RN is functional
	split_functs_methods(Es,FEs,MEs).
split_functs_methods([E|Es],FEs,[E|MEs]) :-
	split_functs_methods(Es,FEs,MEs).


crs_to_list_for_pubs(Eqs) :-
	findall(eq(A,B,D,E_p,F),(builtin_or_abs_eq(A,B,_,D,E,F,_),
			         remove_outputs(E,E_p)),
		Eqs).

process_eqs_for_cc(_,[],[]).
process_eqs_for_cc(CC,[Eq|Eqs],[Eq_p|Eqs_p]) :-
	eq_for_cc(CC,Eq,Eq_p),
	process_eqs_for_cc(CC,Eqs,Eqs_p).


eq_for_cc(main,eq(RN,In,Expr,Calls,Phi),eq(RN,In,Expr,Calls,Phi)) :-
	abs_rule(RN,_,_,_,_,[_,none,_]), !.
eq_for_cc(CC,eq(RN,In,_Expr,Calls,Phi),eq(RN,In,0,Calls_p,Phi)) :-
	abs_rule(RN,_,_,_,_,[_,Class,_]),
	CC \== Class, Class \== none, !,
	remove_funct_calls(Calls,Calls_p).
eq_for_cc(_CC,Eq,Eq).


remove_funct_calls([],[]).
remove_funct_calls([C|Cs],Cs_p) :-
	functor(C,CallRN,_),
	abs_rule(CallRN,_,_,_,_,[_,none,_]), !, % This means CallRN is functional
	remove_funct_calls(Cs,Cs_p).
remove_funct_calls([C|Cs],[C|Cs_p]) :- 
	remove_funct_calls(Cs,Cs_p).


% For debugging
print_eqs(CC,Eqs) :-
	format("~n~nCost center: ~q~n",[CC]),
	member(Eq,Eqs),
	numbervars(Eq,0,_),
	write_term(Eq,[numbervars(true),quoted(true)]), nl,
	fail.
print_eqs(_,_).
