
/** <module> Field assertions

This module provides the interface for the Costabs field
assertions. There are two kinds of assertions. 1) "requires"
annotations provide invariants on the fields used in rules and will be
assumed at their beginnings and therefore conjoined together with
their size relations during abstract compilation. 2) "ensures"
annotations specify properties on the fields after release points,
e.g. they can specify those fields that cannot be modified so that
after the release we can assume the same values the fields had before
the release.
  
@author Miky Zamalloa
@license GPL
*/

:- module(field_assertions,[load_assertions/0, requires/2, ensures/3, all_params/1,
			    parse_await_anns/3, cleanup_field_assertions/0]).

:- use_module(loaded_abs_db,[module/1]).
:- use_module(field_assertions_analysis,[field_assertions_analysis/0]).
:- use_module(stdlib(profiling),[profiling_start_timer/1,profiling_stop_timer/2]).
:- use_module(costabs_output_manager,[info_message/2,write_file/2]).

:- dynamic all_params/1. % Global set of annotation parameters (of the form a(Atom))
:- dynamic requires/2. % requires(RN,Invariant)
:- dynamic ensures/3. % ensures(ReleaseId,UnModifiedFieldsList,Invariant)


cleanup_field_assertions :-
	retractall(all_params(_)),
	retractall(requires(_,_)),
	retractall(ensures(_,_,_)).

load_assertions :- 
	module(Module),
	manual_assertions(Module),
	fail.
load_assertions :- all_params(_), !.
load_assertions :- % If no assertions have been loaded we do the analysis
	profiling_start_timer(f_assers),
	field_assertions_analysis,
	profiling_stop_timer(f_assers,T),
	info_message(f_assers,[T]).


parse_await_anns(CN,Id,Anns) :-
	parse_await_anns_(CN,Anns,UnmodFs,Inv),
	% writeln(ensures(Id,UnmodFs,Inv)),
	((UnmodFs \== []; Inv \== []) -> assertz(ensures(Id,UnmodFs,Inv))
	                               ; true).

parse_await_anns_(CN,[Ann|Anns],[UnmField|UnmFs],Inv) :-
	parse_await_ann(CN,Ann,UnmField,field), !,
	parse_await_anns_(CN,Anns,UnmFs,Inv).
parse_await_anns_(CN,[Ann|Anns],UnmFs,[InvCond|Inv]) :-
	parse_await_ann(CN,Ann,InvCond,inv), 
	parse_await_anns_(CN,Anns,UnmFs,Inv).
parse_await_anns_(_,[],[],[]).

parse_await_ann(_CN,Ann,field(FName),field) :-
	Ann = eq(apply(old(l(FName))),l(FName)).


	
%% By now we write the assertions by hand here

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     Annotations for the PeerToPeer example     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

manual_assertions('PeerToPeer') :-
	assertz(all_params([a(x),a(y)])),
	assertz(requires('DataBaseImpl.getLength',[a(x) >= field(db)])),
	assertz(requires('DataBaseImpl.getFile',[a(x) >= field(db)])),
	assertz(requires('DataBaseImpl.listFiles',[a(x) >= field(db)])),
	assertz(ensures(8,[],[a(x) >= field(db)])), % reqFile.await l2?
	assertz(ensures(6,[],[a(y) >= l(cL)])), % availFiles.await fN? & cL?
	assertz(ensures(1,[],[a(y) >= l(f)])), % run.await f?
	assertz(ensures(2,[],[a(y) >= l(c), a(y) >= field(file),  field(file) >= 0])),  % run.await c?	
	true.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    Annotations for the  MailServer example     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% a(x) -> Number of users in the Mail Server

%% manual_assertions('MailServer') :-
%% 	assertz(all_params([a(x),a(y)])),
%% 	assertz(requires('MailServerImp.notify',[a(x) >= field(listUsers)])),
%% 	assertz(requires('AddressBookImp.getUserAddress',[a(y) >= field(users)])),
%% 	assertz(requires('UserImp.receive',[])),
%% 	assertz(ensures(0,[field(listUsers)],[])), % MailServerImpl.await u
%% 	true.
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    Annotations for the MyPeerToPeer example    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% a(x) -> Number of files in the DB (list of pairs filename/file)
% a(y) -> Number of peers in the P2P network
% a(z) -> Field size of class Node

manual_assertions('MyPeerToPeer') :-
	assertz(all_params([a(x),a(y),a(z)])),
	assertz(requires('DataBaseImpl.getFile',[a(x) >= field(db)])),
	assertz(requires('DataBaseImpl.getLength',[a(x) >= field(db)])),
	assertz(requires('DataBaseImpl.listFiles',[a(x) >= field(db)])),

	assertz(ensures(1,[],[a(y) >= l(f)])), % searchFile.await f?
	assertz(ensures(2,[],[a(y) >= l(c)])),  % searchFile.await c?
	assertz(ensures(7,[],[a(y) >= l(cL)])), % availFiles.await fN? & cL?
	assertz(ensures(8,[field(size)],[field(size) = a(z), a(z)>=1])), % reqFile.await l1?;
	assertz(ensures(9,[field(size)],[a(x) >= l(l2),field(size) = a(z), a(z)>=1])), % reqFile.await l2?;
	true.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%      Annotations for the BookShop example      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

manual_assertions('BookShop') :-
	assertz(all_params([a(x),a(y),a(z)])),
	assertz(requires('AgentImp.free',[a(x) >= field(sessions)])),
	assertz(requires('DatabaseImp.getInfo',[a(y) >= field(booksDB),a(z) >= field(cDB)])),
	assertz(requires('DatabaseImp.confirmOrder',[a(y) >= field(booksDB)])),
	assertz(requires('SyncClient.run',[a(x) >= field(b),a(x) >= field(c)])),
	true.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%        Annotations for the BoundedBuffer       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

manual_assertions('BoundedBuffer') :-
	assertz(all_params([a(x)])),
	assertz(requires('BoundedBuffer.append',[a(x) >= field(buffer)])),
	assertz(ensures(0,[field(buffer)],[a(x) >= field(buffer)])), % BoundedBuffer.await (n < max)
	true.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       Annotations for the DistHT example       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% a(x) -> Number of nodes in the HT
% a(y) -> Maximum number of pairs in a node's map

manual_assertions('DistHT') :-
	assertz(all_params([a(x),a(y)])),
	assertz(requires('DistHT.init',[a(x) >= field(nodes)])),
	assertz(requires('DistHT.lookupNode',[a(x) >= field(nodes),a(x) >= field(nNodes)])),
	assertz(requires('Node.getData',[a(y) >= field(map)])),
	assertz(requires('Node.putData',[a(y) >= field(map)])),
	true.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%          Annotations for other examples        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

manual_assertions('PingPong') :-
	assertz(all_params([])).


