/** <module> Loaded ABS program database

This module provides an interface for the DB in which the current
loaded ABS program is stored. 
  
Currently we only handle the ABS functional sub-part. Therefore, the
DB is a set of facts of the form data/2, type/2 and def/4,
representing, resp., data type declarations, type synonyms and
function definitions.
  
@author Miky Zamalloa and Antonio Flores
@license GPL
*/

:- module(loaded_abs_db,[module/1,
			 type/2, data/2, def/4,
			 interface/3, class/3, method/4, mainBlock/2,
			 add_module/1,
			 add_type/2, add_data/2, add_def/4,
			 add_interface/3, add_class/3, add_method/4, add_mainBlock/2,
			 cleanup_ladb/0, print_ladb/0]).


:- dynamic module/1. % Directives 
:- dynamic type/2, data/2, def/4. % For the functional part
:- dynamic interface/3, class/3, method/4, mainBlock/2. % For the imperative part

%% E.g. given the abs program:
%    type I = Int ;
%    data PNum = Z | S(PNum) ;
%    def PNum max(PNum x, PNum y) =
%      case x > y { True => x ; False => y } ;
%
% the following DB is generated:
%    type('I','Int')
%    data('PNum',['Z','S'('PNum')])
%    def(max,This,'PNum', [('PNum',x),('PNum',y)], case(gt(x,y), [branch('True',x), branch('False',y)]))

add_module(ModuleName):- assertz(module(ModuleName)).

add_type(Name,OriginalName) :- assertz(type(Name,OriginalName)).

add_data(Name,Constructors) :- assertz(data(Name,Constructors)).

add_def(Name,ReturnType,Arguments,Implementation) :- 
	assertz(def(Name,ReturnType,Arguments,Implementation)).

add_interface(Name,Extends,Methods) :- assertz(interface(Name,Extends,Methods)).

%class(name,implementedInterfaces,fieldsList)
add_class(Name,ImplementedInterfaces,Fields):-
	assertz(class(Name,ImplementedInterfaces,Fields)).

add_method(MSig,Class,Anns,Block) :- assertz(method(MSig,Class,Anns,Block)).

add_mainBlock(VarList,Stmts):- assertz(mainBlock(VarList,Stmts)).

cleanup_ladb :-
	retractall(module(_)),
	retractall(type(_,_)),
	retractall(data(_,_)),	
	retractall(def(_,_,_,_)),
	retractall(interface(_,_,_)),
	retractall(class(_,_,_)),
	retractall(method(_,_,_,_)),
	retractall(mainBlock(_,_)).
	

print_ladb :-
	ho_print_pred(module/1),
	ho_print_pred(type/2),
	ho_print_pred(data/2),
	ho_print_pred(def/4),
	ho_print_pred(interface/3),
	ho_print_pred(class/3),
	ho_print_pred(method/4),
	ho_print_pred(mainBlock/2).


ho_print_pred(F/N) :-
	functor(Call,F,N),
	catch(call(Call),error(_,_),fail), % This is to avoid errors when no clause is found
	write(Call),nl,
	fail.
ho_print_pred(_).

