/** <module> The Costabs parameters information DB

This module stores the set of available Costabs parameters as a set of

  param_info(Name,PossibleValues,DefaultValue), where:

Name is the name of a parameter, PossibleValues the list of values it
can take and DefaultValue its default value.
 
@author Miky Zamalloa
@license GPL

*/

:- module(params_info_db,[param_info/3]).


/* There are basically two kinds of parameters: 1) Those that are to
be used externally by the user (most of them available to the web and
eclipse interface, and 2) the internal ones, that are not thought to
be used by the user but for the developers to obtain different
behaviors, e.g. to produce await continuations for aPET)
*/

% 1) External parameters:
% -----------------------

% Cost model. Termination is considered as a cost model.
param_info(cost_model,[termination,steps,memory,objects,task_level,cogs],steps).

% Whether to generate cost centers or not
param_info(cost_centers,[yes,no],no).

% Verbosity level
param_info(verbosity,[0,1,2],2).

% Whether to obtain symbolic or numeric UBs (only applicable for
% "memory" and "objects" cost models)
param_info(symb_or_num,[symb,num],num).

% Size abstraction to be used in functional ABS
param_info(size_abst,[term_size,term_depth],term_size).

% Debug mode
param_info(debug,[yes,no],no).

% Obtain asymptotic bounds
param_info(asymp,[yes,no],no).



% 2) Internal parameters:
% -----------------------

% Term-depth precision. For $match(X,f(Y,Z)), whether to take only {X
% >= Y, X >= Z} or take also {X = 1+max(Y,Z)} i.e. {X = 1+Y} or {X =
% 1+Z}
param_info(td_precision,[1,2],2).

% Enable or not size analysis
param_info(enable_sabu,[yes,no],yes).

% Whether to abstract (reference) fields using path length (as in
% Costa) or just with 0 and 1.
param_info(field_abstraction,[zero_one,path_length],path_length).

% Whether to include the ABS stdlib or not
param_info(stdlib,[yes,no],yes).

% Format of the generated CRS in debug mode
param_info(crs_format,[pretty,pubs],pretty).



% Parameters to tune the generation of the RBR for aPET
% -----------------------------------------------------

% If set to 'yes' field accesses are compiled and represented as
% field(f) variables. Otherwise we keep field accesses as "s(0) =
% field(f)" and field setting as "field(f) = 1". The latter behavior
% is necessary when using the RBR for aPET (for rbr2clp)
param_info(field_vars_in_rbr,[yes,no],yes).


% If set to 'yes': 1) the code after await and get statements will be
% compiled as a continuation in a separate RBR rule. 2) the code that
% evaluates await conditions will be compiled in a separate rule. 3)
% sync calls will be translated to async + get.
param_info(apet_mode,[yes,no],no).

