
/** <module> Generator of the CRS from the RBR(abs)

This module implements the generator of the CRS from the current
RBR(abs) (stored in the rbr_abs_db module). The generated CRS is
stored in module crs_abs_db. The process basically consists in six
steps: 1) We perform some transformations over the RBR, namely,
variable names are converted into actual Prolog variables and we
produce SSA form. 2) Field variables are generated according to the
provided annotations (if no annotation is given we produce fresh field
vars at every release point). 3) Field and annotation vars are
included within the arguments of the corresponding rule calls. 4) The
RBR is traversed and the corresponding CRS is generated just inferring
intra-rules size abstractions. 5) Cost center expressions are
introduced within the generated cost eqs. 6) A bottom-up size analysis
is carried out (module sabu_crs_abs) and the inferred input-output
relations are added to the CRS. Step 6) is optional (depending on
option "enable_sabu").
  
@author Miky Zamalloa
@license GPL
*/

:- module(rbr2crs,[rbr2crs/0, cleanup_rbr2crs/0]).

:- use_module(rbr_abs_db,[abs_rule/6, rule_field/2, build_sccs/0, rule_kind/2, rule_class/2, call_rel/2]).
:- use_module(crs_abs_db,[add_abs_eq/7, add_rels_to_abs_eqs/3]).
:- use_module(sabu_crs_abs,[sabu_fixpoint/0, sabu_memo_table/5]).
:- use_module(costabs_params,[costabs_param/2]).
:- use_module(size_abstractions,[size_abstraction/2, fields_abstraction/5]).
:- use_module(class_invariants,[entry_invariant/2, await_invariant/3, class_inv_params/2]).
:- use_module(costabs_output_manager,[info_message/2]).
:- use_module(cost_models,[apply_cost_model/2]).

:- use_module(stdlib(profiling),[profiling_start_timer/1,profiling_stop_timer/2]).
:- use_module(stdlib(utils),[ut_append_lists/2,ut_append/3,ut_split_atom/3,
			     ut_number_to_atom/2,ut_flat_list/2,ut_sort_rdup/2,
			     ut_nth/3, ut_replace_nth/4]).


rbr2crs :-
	init_rbr2crs, 
	profiling_start_timer(abst_comp),
	abs_rule(RName,This,In,Out,Body,_),

	% Steps 1), 2) and 3)
	vars_ssa_fields(RName,[This|In],Out,Body,
			      [This_p|In_p],Out_p,Body_p,Inv),

	% Steps 4) and 5)
	generate_cost_eq(RName,This_p,In_p,Out_p,Body_p,Inv),
	fail.
rbr2crs :-
	profiling_stop_timer(abst_comp,T_ac),
	info_message(abst_comp,[T_ac]),
	
	% Step 6)
	(costabs_param(enable_sabu,yes) -> 
	    (profiling_start_timer(sabu),
	     build_sccs,
	     sabu_fixpoint,
	     add_sabu_rels,
	     profiling_stop_timer(sabu,T_sabu),
	     info_message(sabu,[T_sabu]))
	;
	     true).
	


add_sabu_rels :-
	sabu_memo_table(F,Args,Rels,_,_),
	add_rels_to_abs_eqs(F,Args,Rels),
	fail.
add_sabu_rels.


generate_cost_eq(RName,This,In,Out,Body,Inv) :-
	process_body(RName,This,Body,Trace,Calls,SizeRelsNoFlat),
	ut_flat_list([Inv,SizeRelsNoFlat],Phi),
	apply_cost_model(Trace,CostExp),
	remove_field_wrappers(In,In_p),	
	(This == none -> EqIn = In_p ; EqIn = [This|In_p]),
	remove_field_wrappers(Out,EqOut),
	new_abs_eq_id(EqId),	
%% 	(costabs_param(cost_centers,yes) ->
%% 	        insert_cost_center(RName,EqIn,CostExp,Calls,EqIn_p,CostExp_p,Calls_p)
%% 	;
%% 	        [EqIn,CostExp,Calls] = [EqIn_p,CostExp_p,Calls_p]),
	% We disable by now the insertion of cost centers
	[EqIn,CostExp,Calls] = [EqIn_p,CostExp_p,Calls_p],
	add_abs_eq(RName,EqIn_p,EqOut,CostExp_p,Calls_p,Phi,[EqId]).

process_body(_,_,[],[],[],[]).
process_body(RN,This,[call(_Kind,RuleCall)|Bs],[call(CallRN)|Es],[Call|Cs],[FAbst|Phis]) :- !,
	RuleCall =..[CallRN,Receiver,In,Out],
	(Receiver == none -> EqIn = In
	                   ; EqIn = [Receiver|In]),
	fields_abstraction(RN,This,[EqIn,Out],[EqIn_p,EqOut],FAbst),
	Call =..[CallRN,EqIn_p,EqOut],
	process_body(RN,This,Bs,Es,Cs,Phis).
process_body(RN,This,[B|Bs],[New,'='|Tr],Cs,[Abst,FAbst|Phis]) :-
	B = '='(V,New),
	nonvar(New), New = new(_CN), !,
	fields_abstraction(RN,This,[V],[V_p],FAbst),
	B_p = '='(V_p,New),
	size_abstraction(B_p,Abst),	
	process_body(RN,This,Bs,Tr,Cs,Phis).
process_body(RN,This,[B|Bs],[New,'='|Tr],Cs,[Abst,FAbst|Phis]) :-
	B = '='(V,New),
	nonvar(New), New = newcog(_CN), !,
	fields_abstraction(RN,This,[V],[V_p],FAbst),
	B_p = '='(V_p,New),
	size_abstraction(B_p,Abst),	
	process_body(RN,This,Bs,Tr,Cs,Phis).
	
process_body(RN,This,[B|Bs],[B|Tr],Cs,[Abst,FAbst|Phis]) :-
	functor(B,_Step,_),
%	member(Step,[match,nomatch,cons,'=',await,release,get]), !,
	fields_abstraction(RN,This,[B],[B_p],FAbst),	
	size_abstraction(B_p,Abst), % Later we flatten the Phis
	process_body(RN,This,Bs,Tr,Cs,Phis).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*
 INSERTION OF COST CENTER EXPRESSIONS WITHIN EQS (Step 5):
  
   a) We add an additional arg to eqs corresponding to functions.
  
   b) Every cost expression is multiplied by the corresponding cost
   center
*/
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

insert_cost_center(RN,EqIn,CostExp,Calls,EqIn_p,CostExp_p,Calls_p) :-
	abs_rule(RN,_,_,_,_,[Kind,none,_]),
	member(Kind,[func,case]), !, % RN is functional
	EqIn_p = [CC|EqIn],
	CCExp = c(cost_center(CC)),
	CostExp_p = CCExp*CostExp,
	add_cc_to_fcalls(CC,Calls,Calls_p).
insert_cost_center(RN,EqIn,CostExp,Calls,EqIn,CostExp_p,Calls_p) :-
	abs_rule(RN,_,_,_,_,[_,Class,_]),
	(Class == none -> CC = 0 % We use the id 0 for the main cost center
	                ; get_cc_mapping(Class,CC)),
	CCExp = c(cost_center(CC)),
	CostExp_p = CCExp*CostExp,
	add_cc_to_fcalls(CC,Calls,Calls_p).


add_cc_to_fcalls(_CC,[],[]).
add_cc_to_fcalls(CC,[Call|RCs],[Call_p|RCs_p]) :-
	Call =..[CallRN,CallIn,CallOut],
	abs_rule(CallRN,_,_,_,_,[_,none,_]), !, % This means CallRN is functional
	Call_p =..[CallRN,[CC|CallIn],CallOut],
	add_cc_to_fcalls(CC,RCs,RCs_p).
add_cc_to_fcalls(CC,[Call|RCs],[Call|RCs_p]) :-
	add_cc_to_fcalls(CC,RCs,RCs_p).


:- dynamic next_cc_id/1.
:- dynamic cc_mapping/2.

init_ccs :-
	assertz(cc_mapping(main,0)),
	assertz(next_cc_id(1)).

inc_next_cc_id :-
	retract(next_cc_id(N)),
	M is N+1,
	assertz(next_cc_id(M)).

get_cc_mapping(Class,CC) :- cc_mapping(Class,CC), !.
get_cc_mapping(Class,CC) :-
	next_cc_id(CC),
	assertz(cc_mapping(Class,CC)),
	inc_next_cc_id.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*
  TRANSFORMATION ON RBR RULES (Steps 1), 2) and 3)):
  
    a) It produces Prolog vars for l and s vars
    
    b) It ensures SSA during step a)
   
    c) It produces Prolog vars for fields in SSA form. Field vars are
    represented as field(V,FN)
  
    d) We use fresh field vars after release points unless an
    assertion indicates not to do so
  
    e) Input and output args are extended so that they include
    annotation params and fields (if so)
  
    f) In case there are entry or await annotations the corresponding
    invariant is processed (var names transformations) and returned
                                                        
*/
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


vars_ssa_fields(RN,[This|In],Out,Body,[This_p|In_p],Out_p,Body_p,Inv) :-
	init_dict,
	% the dictionary is a difference list in which I need a dummy element '0'
	Dict = [0|DictTail],
	extend_input_args(RN,In,ExtIn),
	extend_output_args(RN,Out,ExtOut),
	load_entry_invariant(RN,ReqInv,(Dict,DictTail),DT_p),
	varnames_to_prologvars([This|ExtIn],[This_p|In_p],(Dict,DT_p),DT_pp),
	body_vars_ssa_fields(Body,Body_p,(Dict,DT_pp),(Dict_p,DT_ppp),ReleaseInvs),	
	varnames_to_prologvars(ExtOut,Out_p,(Dict_p,DT_ppp),_),
	ut_flat_list([ReqInv,ReleaseInvs],Inv),
	cleanup_dict.


varnames_to_prologvars([],[],(_,DT),DT).
varnames_to_prologvars([V|Ts],[V|Ts_p],(Dict,DictTail),DT_p) :-
	var(V), !,
	info_message(found_variable_error,[]),
	varnames_to_prologvars(Ts,Ts_p,(Dict,DictTail),DT_p).
varnames_to_prologvars(['_'|Ts],[_|Ts_p],(Dict,DictTail),DT_p) :- !,
	varnames_to_prologvars(Ts,Ts_p,(Dict,DictTail),DT_p).
varnames_to_prologvars([VName|Ts],[V_p|Ts_p],(Dict,DictTail),DT_p) :-
	member(VName,[l(_),s(_),field(_),max(_),min(_)]),
	dict_index(VName,I), !, % VName is already in the dictionary
	ut_nth(I,Dict,V),
	(VName = field(FN) -> V_p = field(V,FN) ; V_p = V),	
	varnames_to_prologvars(Ts,Ts_p,(Dict,DictTail),DT_p).
varnames_to_prologvars([VName|Ts],[V_p|Ts_p],(Dict,DictTail),DT_pp) :-
	% VName is not in the dictionary
	member(VName,[l(_),s(_),field(_),max(_),min(_)]), !,
	dict_count(I), 
	assertz(dict_index(VName,I)),
	inc_dict_count,
	DictTail = [V|DictTail_p],
	(VName = field(FN) -> V_p = field(V,FN) ; V_p = V),		
	varnames_to_prologvars(Ts,Ts_p,(Dict,DictTail_p),DT_pp).
varnames_to_prologvars([T|Ts],[T_p|Ts_p],(Dict,DictTail),DT_pp) :-
	compound(T), !,
	T =..[F|Args],
	varnames_to_prologvars(Args,Args_p,(Dict,DictTail),DT_p),
	T_p =..[F|Args_p],
	varnames_to_prologvars(Ts,Ts_p,(Dict,DT_p),DT_pp).
varnames_to_prologvars([T|Ts],[T|Ts_p],(Dict,DictTail),DT_p) :-
	% T is not '_', nor a var, nor an l, s or field var neither compound
	varnames_to_prologvars(Ts,Ts_p,(Dict,DictTail),DT_p).


body_vars_ssa_fields([],[],Dict,Dict,[]).
body_vars_ssa_fields([B|Bs],[B_p|Bs_p],(Dict,DT),OutDict,[Inv|Invs]) :-
	b_vars_ssa_fields(B,B_p,(Dict,DT),(Dict_p,DT_p),Inv),
	body_vars_ssa_fields(Bs,Bs_p,(Dict_p,DT_p),OutDict,Invs).

b_vars_ssa_fields(AssignStmt,AssignStmt_p,(Dict,DT),(Dict_p,DT_pp),[]) :-
	AssignStmt =..['=',VName,Exp], !,
	varnames_to_prologvars([Exp],[Exp_p],(Dict,DT),DT_p),
	(VName = field(FN) -> V_p = field(V,FN) ; V_p = V),	
	(dict_index(VName,I) -> 
	    (ut_replace_nth(Dict,I,V,Dict_p),
	     DT_pp = DT_p)
	; 
	    (dict_count(I), assertz(dict_index(VName,I)), inc_dict_count,
	     DT_p = [V|DT_pp],
	     Dict_p = Dict)),
	AssignStmt_p =..['=',V_p,Exp_p].

% If release point -> take assertion and produce fresh fieldvars accordingly
b_vars_ssa_fields(ReleaseOrAwait,ReleaseOrAwait_p,(Dict,DT),(Dict_p,DT_pp),Inv) :-
	ReleaseOrAwait =..[RorA,Id|G], member(RorA,[release,await]), !,
	(await_invariant(Id,UnmFs,_) -> true ; UnmFs = []),
	findall(I, (dict_index(field(F),I), \+ member(field(F),UnmFs)),
		Is), % Is should be sorted
	length(Is,Length),
	length(FreshVars,Length),
	replace_sorted_nths(Dict,Is,FreshVars,Dict_p),
	varnames_to_prologvars([G],[G_p],(Dict_p,DT),DT_p),
	ReleaseOrAwait_p =..[RorA|G_p],
	load_await_invariant(Id,Inv,(Dict_p,DT_p),DT_pp).

% If call to if, while or cont -> add the corresponding field vars (do
% not forget to add them also in rule heads)
b_vars_ssa_fields(call(Kind,RuleCall),Call_p,(Dict,DT),(Dict_p,DT_pp),[]) :-
	RuleCall =..[RN,Receiver,In,Out],
	extend_input_args(RN,In,In_p),
	extend_output_args(RN,Out,Out_p),	
	varnames_to_prologvars([Receiver,In_p],
			       [Receiver_p,In_pp],(Dict,DT),DT_p),
	% We must produce fresh vars for output vars (of while's) to ensure SSA
	fresh_out_vars(Out_p,Out_pp,(Dict,DT_p),(Dict_p,DT_pp)),
	RuleCall_p =..[RN,Receiver_p,In_pp,Out_pp],
	Call_p = call(Kind,RuleCall_p).

b_vars_ssa_fields(Stmt,Stmt_p,(Dict,DT),(Dict,DT_p),[]) :-
	Stmt =..[Inst|Args],
	varnames_to_prologvars(Args,Args_p,(Dict,DT),DT_p),
	Stmt_p =..[Inst|Args_p].


fresh_out_vars([],[],(Dict,DT),(Dict,DT)).
fresh_out_vars([VName|VTs],[V|Vs],(Dict,DT),(Dict_pp,DT_pp)) :-
	(dict_index(VName,I) -> 
	        (ut_replace_nth(Dict,I,V,Dict_p),
		 DT_p = DT)
	; 
	        (dict_count(I), assertz(dict_index(VName,I)), inc_dict_count,
		 DT = [V|DT_p],
		 Dict_p = Dict)),
	fresh_out_vars(VTs,Vs,(Dict_p,DT_p),(Dict_pp,DT_pp)).
	

%% We extend the input args with the fields (only for intermediate blocks)
%  and the annotations parameters (except for functions and cases).
extend_input_args(RN,In,In_ppp) :-
	rule_kind(RN,Kind),
	(is_imp_block_kind(Kind) ->
	        (add_field_vars(In,RN,In_p),
		 add_annot_params(_,In_p,In_pp))
	;
	        In_pp = In),
	(Kind = method -> add_annot_params(RN,In_pp,In_ppp)
	                ; In_ppp = In_pp).

extend_output_args(RN,Out,Out_p) :-
	rule_kind(RN,Kind),
	(is_imp_block_kind(Kind) ->
	        add_field_vars(Out,RN,Out_p)
	;
	        Out_p = Out).

add_field_vars(In,RN,InAndFs) :-
	% We compute the field vars for RN only if it has not been already computed
	(field_vars(RN,Fs) -> true ; compute_field_vars(RN,Fs)),
	append(In,Fs,InAndFs).

add_annot_params(_RN,In,In_p) :-
	%all_params(Params), !,
	%rule_class(RN,CN),
	class_inv_params(_CN,Params), 
	append(In,Params,In_p).
add_annot_params(_,In,In).

load_entry_invariant(RN,Inv_pp,(Dict,DT),DT_p) :-
	entry_invariant(RN,Inv), !,
	varnames_to_prologvars(Inv,Inv_p,(Dict,DT),DT_p),
	remove_field_wrappers(Inv_p,Inv_pp).
load_entry_invariant(_,[],(_,DT),DT).

load_await_invariant(Id,Inv_pp,(Dict,DT),DT_p) :- 
	await_invariant(Id,_,Inv), !,
	varnames_to_prologvars(Inv,Inv_p,(Dict,DT),DT_p),
	remove_field_wrappers(Inv_p,Inv_pp).
load_await_invariant(_,[],(_,DT),DT).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                VARS DICTIONARY                 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% The vars dictionary is a list of variables (represented as a
%  difference list) which can be seen as an array. We store the
%  mappings between var names and positions in the list (or array) as
%  dict_index/2 facts. 

:- dynamic dict_count/1. % Number of elements in the dictionary
:- dynamic dict_index/2. % dict_index(Name,I) var Name is at index I of the dictionary

init_dict :-
	cleanup_dict,
	assertz(dict_count(2)). % we need a starting dummy element to init the diff list

cleanup_dict :-
	retractall(dict_index(_,_)),
	retractall(dict_count(_)).

inc_dict_count :-
	retract(dict_count(N)),
	M is N+1,
	assertz(dict_count(M)).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                AUXILIARY PREDICATES                 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


is_imp_block_kind(Kind) :- member(Kind,[if,cont,while,entrywhile,virtual,awguard]).


%% To store transitively used field vars per rule (only for imperative block rules)
:- dynamic field_vars/2.
				
%% This predicate computes the set of fields which are used by rule
%  RN, including those used in calls to other intermediate imperative
%  rules (transitively). Only applicable if the rule is an imperative
%  block rule.
compute_field_vars(RN,Fs_p) :-
	rule_kind(RN,Kind),
	is_imp_block_kind(Kind),
	findall(F, field_var_trans(RN,F,[RN]), Fs),
	ut_sort_rdup(Fs,Fs_p),
	%info_message(field_vars_info,[field_vars(RN,Fs_p)]),
	assertz(field_vars(RN,Fs_p)).

field_var_trans(RN,field(F),_) :- rule_field(RN,F).
field_var_trans(RN,F,L) :-
	call_rel(RN,RN_p),
	\+ member(RN_p,L), % to control loops
	rule_kind(RN_p,Kind_p),
	is_imp_block_kind(Kind_p),
	field_var_trans(RN_p,F,[RN_p|L]).


% It assumes that the indexes list is sorted
replace_sorted_nths(L,Is,Vs,L_p) :- replace_sorted_nths_(L,Is,Vs,L_p,1).

replace_sorted_nths_(R,_,_,R,_) :- var(R),!.
replace_sorted_nths_(L,[],[],L,_).
replace_sorted_nths_([_|Xs],[I|Is],[V|Vs],[V|Xs_p],I) :- !,
	I_p is I + 1,
	replace_sorted_nths_(Xs,Is,Vs,Xs_p,I_p).
replace_sorted_nths_([X|Xs],[I|Is],[V|Vs],[X|Xs_p],CurrI) :-
	CurrI_p is CurrI + 1,
	replace_sorted_nths_(Xs,[I|Is],[V|Vs],Xs_p,CurrI_p).


remove_field_wrappers([],[]).
remove_field_wrappers([V|Ts],[V|Ts_p]) :-
	var(V), !,
	remove_field_wrappers(Ts,Ts_p).
remove_field_wrappers([field(V,_)|Ts],[V|Ts_p]) :- !,
	remove_field_wrappers(Ts,Ts_p).
remove_field_wrappers([T|Ts],[T_p|Ts_p]) :-
	T =..[F|Args],
	remove_field_wrappers(Args,Args_p),
	T_p =..[F|Args_p],
	remove_field_wrappers(Ts,Ts_p).


:- dynamic abs_eq_id/1.

init_rbr2crs :-
	cleanup_rbr2crs,
	assertz(abs_eq_id(0)),
	init_ccs.

new_abs_eq_id(Id) :-
	retract(abs_eq_id(Id)),
	Id_p is Id+1,
	assertz(abs_eq_id(Id_p)).

cleanup_rbr2crs :-
	retractall(field_vars(_,_)),
	retractall(abs_eq_id(_)),
	retractall(next_cc_id(_)).	

