/** <module> RBR(abs) Database

This module provides an interface for the DB in which the current
RBR(abs) representing the current ABS program is stored. Besides the
RBR we store additional information like: rule call relations, used
fields within a rule and SCC's.
  
@author Miky Zamalloa
@license GPL
*/


:- module(rbr_abs_db,[abs_rule/6, add_abs_rule/6,
		      builtin_or_abs_rule/6, replace_rule_body/3,
		      call_rel/2, add_call_rel/2,
		      build_sccs/0, rbr_scc/3,
		      rule_field/2, add_rule_field/2,
		      rule_kind/2, rule_class/2, field_type/3,
		      print_rbrdb/0,cleanup_rbrdb/0]).


:- use_module(stdlib(scc),[compute_sccs/2]).
:- use_module(loaded_abs_db,[def/4,class/3]).


%% The RBR(abs) is represented as a set of facts
%
%          abs_rule(Name,This,InArgs,OutArgs,Body,Info)
%
%  where Body is a list of terms of the following form:
%    * V = W, where V can be a var or field(F), and W a var, a field(F), or a literal
%    * match(V,ConsPattern) and nomatch(V,ConsPattern)
%    * cons(V,ConsPattern)
%    * call(Kind,r(This,InArgs,OutArg)) -> rule calls, including arithmetic builtins
%                                          Kind \in {sync,async,func}
%    * await(Id,G) -> The Id is global in the whole RBR
%    * release(Id) -> It corresponds to the ABS "suspend"
%    * native  -> function definition for which there is no code
%    * $op(X,Y) -> Arithmetic builtins (only in arithmetic functs: eq, leq, etc)
%
%  and Info a list of additional tokens of information, currently of
%  the form [kind,Class,RuleId] with Kind \in
%  {func,case,method,if,cont,entrywhile,while,virtual,awguard}, Class the
%  class from where the rule originates (or none if it is a functional
%  rule or the main), and, RuleId an unique identifier for the rule.
%
% Additionally we store the call graph as a set of call_rel/2 facts,
% the fields which are used within a rule (as rule_field/2 facts),
% and, if requested (before calling the size analysis), the set of
% SCCs stored as a set of rbr_scc/3 facts.

:- dynamic abs_rule/6.
:- dynamic call_rel/2.
:- dynamic rule_field/2.
:- dynamic rbr_scc/3.

add_abs_rule(A,B,C,D,E,F) :- assertz(abs_rule(A,B,C,D,E,F)).

%% This replaces the body of rule RId but also moves to the end of the
%  DB rules with name RN so that they remain together in the RBR.
replace_rule_body(RN,RId,NewBody) :- 
	retract(abs_rule(RN,This,In,Out,Body,[Kind,CN,RId_p])),
	(RId == RId_p -> Body_p = NewBody ; Body_p = Body),
	assertz(abs_rule(RN,This,In,Out,Body_p,[Kind,CN,RId_p])),
	fail.
replace_rule_body(_,_,_).

add_call_rel(A,B) :- call_rel(A,B), !.
add_call_rel(A,B) :- assertz(call_rel(A,B)).

add_rule_field(R,F) :- rule_field(R,F), !.
add_rule_field(R,F) :- assertz(rule_field(R,F)).

%build_sccs :-
%	assert(rbr_scc(0,['Node.availFiles',if_1,cont_1,slhead,case_35'Node.enquire',sltail,slappendright],recursive)),
% 	assert(rbr_scc(1,[case_11,domain],recursive)), !.
build_sccs :-
	findall(A-B, call_rel(A,B), Graph),
	compute_sccs(Graph,SCCs),
	assert_sccs(SCCs,0).
%	nl, print_sccs.

assert_sccs([],_).
assert_sccs([(Type,Nodes)|SCCs],N) :-
	assertz(rbr_scc(N,Nodes,Type)),
	N_p is N+1,
	assert_sccs(SCCs,N_p).

rule_kind(RN,Kind) :- abs_rule(RN,_,_,_,_,[Kind,_,_]), !.

rule_class(RN,Class) :- abs_rule(RN,_,_,_,_,[_,Class,_]), !.


field_type(RN,FN,Type) :-
	rule_class(RN,CN),
	\+ CN = none, !,
	class(CN,_,Fields),
	member(field(FN,Type,_),Fields).

builtin_or_abs_rule(F,This,In,Outs,Bs,Info) :- builtin_rule(F,This,In,Outs,Bs,Info).
builtin_or_abs_rule(F,This,In,Outs,Bs,Info) :- abs_rule(F,This,In,Outs,Bs,Info).

% Builtin-rules for relational operators
builtin_rule(eq,none,[l(x),l(y)],[l(z)],[eq(l(x),l(y)), l(z) = 'True'],[func,none,b1]).
builtin_rule(eq,none,[l(x),l(y)],[l(z)],[neq(l(x),l(y)), l(z) = 'False'],[func,none,b2]).
builtin_rule(neq,none,[l(x),l(y)],[l(z)],[neq(l(x),l(y)), l(z) = 'True'],[func,none,b3]).
builtin_rule(neq,none,[l(x),l(y)],[l(z)],[eq(l(x),l(y)), l(z) = 'False'],[func,none,b4]).
builtin_rule(lt,none,[l(x),l(y)],[l(z)],[lt(l(x),l(y)), l(z) = 'True'],[func,none,b5]).
builtin_rule(lt,none,[l(x),l(y)],[l(z)],[geq(l(x),l(y)), l(z) = 'False'],[func,none,b6]).
builtin_rule(leq,none,[l(x),l(y)],[l(z)],[leq(l(x),l(y)), l(z) = 'True'],[func,none,b7]).
builtin_rule(leq,none,[l(x),l(y)],[l(z)],[gt(l(x),l(y)), l(z) = 'False'],[func,none,b8]).
builtin_rule(gt,none,[l(x),l(y)],[l(z)],[gt(l(x),l(y)), l(z) = 'True'],[func,none,b9]).
builtin_rule(gt,none,[l(x),l(y)],[l(z)],[leq(l(x),l(y)), l(z) = 'False'],[func,none,b10]).
builtin_rule(geq,none,[l(x),l(y)],[l(z)],[geq(l(x),l(y)), l(z) = 'True'],[func,none,b11]).
builtin_rule(geq,none,[l(x),l(y)],[l(z)],[lt(l(x),l(y)), l(z) = 'False'],[func,none,b12]).
builtin_rule(neg,none,[l(x)],[l(z)],[eq(l(x),'True'), l(z) = 'False'],[func,none,b13]).
builtin_rule(neg,none,[l(x)],[l(z)],[eq(l(x),'False'), l(z) = 'True'],[func,none,b14]).
%builtin_rule(and,none,[l(x),l(y)],[l(z)],[eq(l(x),'False'), l(z) = 'False'],[func,none,b15]).
%builtin_rule(and,none,[l(x),l(y)],[l(z)],[eq(l(x),'True'), l(z) = l(y)],[func,none,b16]).


from_functional_abs(Name) :- def(Name,_,_,_),!.
from_functional_abs(Name) :- call_rel(F,Name), from_functional_abs(F).


cleanup_rbrdb :-
	retractall(abs_rule(_,_,_,_,_,_)),
	retractall(call_rel(_,_)),
	retractall(rule_field(_,_)),
	retractall(rbr_scc(_,_,_)).

print_rbrdb :-
	format('~n~n-----------------------~n',[]),
	format('        RBR(abs)       ~n',[]),
	format('-----------------------~n',[]),
	builtin_or_abs_rule(F,This,In,Out,Bs,_),
	numbervars([F,This,In,Out,Bs],0,_),
	format('~n~q(',[F]),
	write_term(This,[numbervars(true),quoted(true)]),
	write(','),	
	write_term(In,[numbervars(true),quoted(true)]),
	write(','),
	write_term(Out,[numbervars(true),quoted(true)]),
	format(') => ~n',[]),
	member(B,Bs),
	format('    ',[]),
	write_term(B,[numbervars(true),quoted(true)]),
	format(',~n',[]),
	fail.
print_rbrdb :-
	nl.
%	print_call_rels.

print_call_rels :-
	call_rel(A,B),
	format('call_rel(~q,~q)~n',[A,B]),
	fail.
print_call_rels.

print_rule_fields :-
	rule_field(A,B),
	format('rule_field(~q,~q)~n',[A,B]),
	fail.
print_rule_fields.

print_sccs :-
	rbr_scc(A,B,C),
	format('rbr_scc(~q,~q,~q)~n',[A,B,C]),
	fail.
print_sccs.
