
/** <module> Bottom-up Size Analysis on the CRS(abs)

This module implements a bottom-up size analysis on the
CRS(abs). Predicate sabu_fixpoint/0 performs the analysis from the set
of SCC's and RBR rules. The resulting size relations are stored as a
set of sabu_memo_table/5 facts.
  
@author Samir Genaim and Miky Zamalloa
@license GPL
*/


:- module(sabu_crs_abs,[sabu_fixpoint/0,sabu_memo_table/5]).


:- use_module(crs_abs_db,[abs_eq/7]).
:- use_module(rbr_abs_db,[rbr_scc/3]).
:- use_module(stdlib(numeric_abstract_domains)).
:- use_module(stdlib(utils),[ut_list_to_dlist/2, ut_member/2, ut_var_member/2,
			     ut_append/3]).


:- dynamic memo_table_changed/0.
:- dynamic sabu_memo_table/5.


sabu_fixpoint :-
	nad_set_domain(c_polyhedra_ppl), % TODO perhaps this should be removed
	nad_initialize, 
	retractall(memo_table_changed),
	retractall(sabu_memo_table(_,_,_,_,_)),
	iterate_scc(0).


iterate_scc(N) :-
	rbr_scc(N,SccRules,SccFlag),
	iterate_tp(SccFlag,SccRules, 0),
	!,
	N1 is N+1,
	iterate_scc(N1).
iterate_scc(_N).


iterate_tp(_SccFlag  ,SccRules, N) :- 
	tp(SccRules, N).
iterate_tp(SccFlag  ,SccRules, N) :- 
	retract(memo_table_changed), 
	SccFlag=recursive,
	N1 is N+1,
	iterate_tp(SccFlag,SccRules,N1).
iterate_tp(_SccFlag ,_SccRules, _N).


tp(SccRules, N) :-
	ut_member(EqName,SccRules),
	abs_eq(EqName,In,Out,_,Calls,Phi,_),

	once(select_call_from_prev_iteration(N,Calls,Calls_1,Cs_0)),
	once(process_calls(Calls_1,Cs_0,Cs_1)),

	ut_append(Cs_1,Phi,Cs_2),
	ut_append(In,Out,Args),
	cond_assert(EqName,Args,Cs_2,N),
	fail.


%%%%
%%%% UNCOMMENT THIS TO DISABLE THIS OPTIMIZATION
%%%%
%select_call_from_prev_iteration(_,Body,Body,[]) :- !.

select_call_from_prev_iteration(N,Body,Body_aux,Cs) :-
	N == 0,
	Body = Body_aux,
	Cs = [],
	!.
select_call_from_prev_iteration(N,Body,Body_aux,Cs) :-
	N>0,
	select_call_from_prev_iteration_aux(Body,N,Body_aux,Cs).

select_call_from_prev_iteration_aux([B|Bs],N,Bs,Cs) :-
	B =..[EqName,In,Out],
	ut_append(In,Out,Args),
	sabu_memo_table(EqName,Args,Cs,N1,_),
	N1 is N-1,
	!.
select_call_from_prev_iteration_aux([B|Bs],N,[B|Rs],Cs) :-
	select_call_from_prev_iteration_aux(Bs,N,Rs,Cs).



%% solve_body([],Cs,Cs).
%% solve_body([B|Bs],Cs_Accm,Cs) :-
%% 	solve_one_item(B,Cs_B),
%% 	ut_append(Cs_B,Cs_Accm,Cs_Accm_aux),
%% 	solve_body(Bs,Cs_Accm_aux,Cs).
%% 
%% %% If it is a constraint we just add it to the store 
%% solve_one_item(size_constraints(Cs),Cs).
%% 
%% %% If it is call we try to get the current answer
%% solve_one_item(call(EqName,Args),Cs) :-
%% 	sabu_memo_table(EqName,Args,Cs,_,_).


process_calls([],Cs,Cs).
process_calls([Call|Calls],Cs_Accm,Cs) :-
	Call =..[F,In,Out],
	ut_append(In,Out,Args),
	sabu_memo_table(F,Args,Cs_B,_,_),
	ut_append(Cs_B,Cs_Accm,Cs_Accm_aux),
	process_calls(Calls,Cs_Accm_aux,Cs).



cond_assert(EqName, Args, Cs, N) :-
	normalize_args(Args, Args_1, Cs, Cs_0),
	nad_project(Args_1,Cs_0,Cs_1),
	(
	 sabu_memo_table(EqName, Args_1, Cs_Old, _, Widen_k) ->
	   nad_lub(Args_1, Cs_Old, Args_1, Cs_1, Args_1, Cs_New),
	   \+ nad_entails(Args_1, Cs_New, Cs_Old),
	   ( Widen_k >= 3 ->
	       nad_widen(Args_1,Cs_Old,Cs_New,Args_1,Cs_Final),
	       Widen_k_aux = 0
	   ;
	       Cs_Final = Cs_New,
	       Widen_k_aux is Widen_k+1
	   ),
	   retractall(sabu_memo_table(EqName, Args_1, _, _, _))
	;
	 Cs_Final = Cs_1,
	 Widen_k_aux = 0
	),
	assertz(sabu_memo_table(EqName, Args_1, Cs_Final, N, Widen_k_aux)),
    	raise_memo_table_flag.

raise_memo_table_flag :-
	memo_table_changed,
	!.
raise_memo_table_flag :-
%	write(a),nl,
	assertz(memo_table_changed).




%% This removes duplicates and numbers, and adds them as explicit
%  constraints.
normalize_args([],[],Cs,Cs).
normalize_args([X|Xs],[V|Ys],Cs_Curr,[V=X|Cs]) :-
	number(X),
	!,
	normalize_args(Xs,Ys,Cs_Curr,Cs).
normalize_args([X|Xs],[V|Ys],Cs_Curr,[V=X|Cs]) :-
	once(ut_var_member(X,Xs)),
	!,
	normalize_args(Xs,Ys,Cs_Curr,Cs).
normalize_args([X|Xs],[X|Ys],Cs_Curr,Cs) :-
	normalize_args(Xs,Ys,Cs_Curr,Cs).



%rbr_scc(0,[case_0,domain],recursive).
%rbr_scc(1,[case_1,contains],recursive).
%rbr_scc(2,[case_22,case_23,getBooks],recursive).

% Example of an abstract program
/*
rbr_scc(0,[case_11,domain],recursive).

rbr_rule(case_11,[A,B],[size_constraints([A=1, B=1])]).
rbr_rule(case_11,[A,B],[size_constraints([A>=1,B>=1,C>=1,D>=1, E>=1,F>=1,A>=3, A-E>=2, A-F>=2, A-C>=1, B>=2, B=E+1]),
			  call(domain,[C, D])]).
rbr_rule(case_11,[A,B],[size_constraints([A>=1,B>=1,C>=1,D>=1, E>=1,F>=1,A>=3, A-E>=2, A-F>=2, A-C>=1, B>=2, B=D+1]),
			  call(domain,[C, D])]).
rbr_rule(domain,[A, B],[call(case_11,[A, B])]).
*/

