
/** <module> Size abstractions

This module implements the size abstractions of Costabs. Given an RBR
statement, first, predicate fields_abstraction/5 produces the field
abstractions on the fields within the statement w.r.t This (if
parameter "field_abstraction" is set to path_length), and generates a
new statement in which the field(FV,FN) terms are replaced by
FV. Then, predicate size_abstraction/2 is applied over the resulting
statement returning in the second argument the list of constraints
defining the corresponding size abstractions. Parameter "size_abst"
specifies the chosen size abstraction. Currently we have two: the term
depth and the term-size.
  
@author Miky Zamalloa
@license GPL
*/

:- module(size_abstractions,[size_abstraction/2, fields_abstraction/5]).

:- use_module(loaded_abs_db,[interface/3]).
:- use_module(rbr_abs_db,[field_type/3]).
:- use_module(costabs_params,[costabs_param/2]).
:- use_module(stdlib(utils),[ut_max/3,ut_varset/2]).
:- use_module(costabs_output_manager,[info_message/2]).
:- use_module(auxiliary_preds,[is_relational_exp/1]).


%% First we have the size abstractions that are common for both
%  term-depth and term-size. If none of them aplies then we look for
%  the specific size abstractions, otherwise the term is not supported

size_abstraction(match(V,BoolOrN),[V = S]) :-
	nonvar(BoolOrN), atom_abstraction(BoolOrN,S), !.
size_abstraction(cons(V,Bool),[V = S]) :- atom_abstraction(Bool,S), !.
size_abstraction('='(X,Y),[X = AY]) :- atom_abstraction(Y,AY), !.
size_abstraction('='(X,mod(_,Z)),[AZ >= X+1]) :- atom_abstraction(Z,AZ), !.
size_abstraction('='(X,Y),[X = AY]) :- arith_abstraction(Y,AY), !.
size_abstraction('='(V,new(_)),[V >= 1]) :- costabs_param(field_abstraction,path_length), !.
size_abstraction('='(V,new(_)),[V = 0]) :- !.
size_abstraction('='(V,newcog(_)),[V >= 1]) :- costabs_param(field_abstraction,path_length), !.
size_abstraction('='(V,newcog(_)),[V = 0]) :- !.
size_abstraction(eq(X,Y),[AX = AY]) :- atom_abstraction(X,AX), atom_abstraction(Y,AY), !.
size_abstraction(neq(X,Y),[X >= 1]) :- Y == null, !. % Special case for X != null
size_abstraction(neq(X,Y),[AX >= AY+1]) :- atom_abstraction(X,AX), atom_abstraction(Y,AY).
size_abstraction(neq(X,Y),[AY >= AX+1]) :- atom_abstraction(X,AX), atom_abstraction(Y,AY), !.
size_abstraction(lt(X,Y),[Y >= X+1]) :- !.
size_abstraction(leq(X,Y),[Y >= X]) :- !.
size_abstraction(gt(X,Y),[X >= Y+1]) :- !.
size_abstraction(geq(X,Y),[X >= Y]) :- !.
size_abstraction(await(G),Phi) :- !, await_guard_abstraction(G,Phi).
size_abstraction(B,Abst) :- 
	(costabs_param(size_abst,term_depth) -> size_abstraction_td(B,Abst)
	                                      ; size_abstraction_ts(B,Abst)).


%% Specific size abstractions for the term-depth norm

%% size_abstraction_td(nomatch(V,BoolOrNum),Abst) :-
%% 	bool_or_num(BoolOrNum), !,
%% 	nomatch_abstraction_td(BoolOrNum,V,Abst).
size_abstraction_td(nomatch(_V,_BoolOrNum),[]) :- !.
size_abstraction_td(match(V,CP),Abst) :- !, % Here CP cannot be a boolean
	normal_term_depth(CP,(Rel,Val)), % (Rel,Val) can be e.g. (>=,2)
	Phi1 =..[Rel,V,Val], 
	term_vars_and_depths(CP,CPVarsAndDepths),
	generate_var_rels(V,CPVarsAndDepths,VarRels),
	(costabs_param(td_precision,2), CPVarsAndDepths = [_|_] -> (
	    member((Var,Depth),CPVarsAndDepths),
	    VarRels_p = [V = Depth+Var|VarRels])
	;
	    VarRels_p = VarRels),
	Abst = [Phi1|VarRels_p].
size_abstraction_td(cons(V,ConsExp),Abst) :- !,
	normal_term_depth(ConsExp,(Rel,Val)), % (Rel,Val) can be e.g. (>,2)
	term_vars_and_depths(ConsExp,CEVarsAndDepths),
	generate_var_rels(V,CEVarsAndDepths,VarRels),
	Phi1 =..[Rel,V,Val], 
	Abst = [Phi1|VarRels].
size_abstraction_td(B,[]) :-
	info_message(not_supported_size_abstraction,[B]).


%% Specific size abstractions for the term-size norm

size_abstraction_ts(nomatch(_V,_BoolOrNum),[]) :- !.
size_abstraction_ts(match(V,CP),[V = TS|Phi]) :- !, % Here CP cannot be a boolean
	ut_varset(CP,CPVs),
	positive_sizes_phi(CPVs,Phi),
	term_size(CP,TS).
size_abstraction_ts(cons(V,ConsExp),[V = TS]) :- !,
	term_size(ConsExp,TS).
size_abstraction_ts(B,[]) :-
	info_message(not_supported_size_abstraction,[B]).



%% fields_abstraction(RN,This,Ts,T_p,Phi) produces the field
%  abstractions Phi on the fields within Ts w.r.t This (if
%  field_abstraction = path_length), and returns Ts_p, in which the
%  field(FV,FN) terms are replaced by FV.
fields_abstraction(_,_,[],[],[]).
fields_abstraction(RN,This,[T|Ts],[T|Ts_p],Phis) :-
	var(T), !,
	fields_abstraction(RN,This,Ts,Ts_p,Phis).
fields_abstraction(RN,This,[field(FV,FN)|Ts],[FV|Ts_p],[This>=FV+1|Phis]) :-
	costabs_param(field_abstraction,path_length),
	field_type(RN,FN,FType), % The path length only applies if the 
	interface(FType,_,_), !, % field is of interface type
 	fields_abstraction(RN,This,Ts,Ts_p,Phis).
fields_abstraction(RN,This,[field(FV,_)|Ts],[FV|Ts_p],Phis) :- !,
	% field_abstraction param is set to zero_one
	fields_abstraction(RN,This,Ts,Ts_p,Phis).
fields_abstraction(RN,This,[T|Ts],[T_p|Ts_p],[ArgsPhi|Phis]) :-
	T =..[F|Args],
	fields_abstraction(RN,This,Args,Args_p,ArgsPhi),
	T_p =..[F|Args_p],
	fields_abstraction(RN,This,Ts,Ts_p,Phis).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                Auxiliary predicates                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

atom_abstraction(X,X) :- var(X), !.
atom_abstraction(null,0).
atom_abstraction('True',1).
atom_abstraction('False',0).
atom_abstraction(N,N) :- number(N).
atom_abstraction(get(X),X).
atom_abstraction(str(_),1).

arith_abstraction(X+Y,X+Y).
arith_abstraction(X-Y,X-Y).
arith_abstraction(-X,-X).
arith_abstraction(_*_,_).
arith_abstraction(_/_,_).
arith_abstraction(mod(_,_),_).

bool_or_num('True').
bool_or_num('False').
bool_or_num(N) :- number(N).

nomatch_abstraction('True',V,[V = 0]).
nomatch_abstraction('False',V,[V = 1]).
nomatch_abstraction(N,V,[V >= N+1]) :- number(N).
nomatch_abstraction(N,V,[N >= V+1]) :- number(N).

await_guard_abstraction(G,[]) :- var(G), !.
await_guard_abstraction(G,Phi) :-
	is_relational_exp(G), !,
	size_abstraction(G,Phi).
await_guard_abstraction(and(G1,G2),Phi) :-
	await_guard_abstraction(G1,Phi1),
	await_guard_abstraction(G2,Phi2),
	append(Phi1,Phi2,Phi).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Auxiliary predicates for the term-depth abstraction  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

positive_sizes_phi([],[]).
positive_sizes_phi([V|Vs],[V >= 1|Phi]) :-
	positive_sizes_phi(Vs,Phi).

generate_var_rels(_V,[],[]).
generate_var_rels(Var,[(V,D)|VDs],[Var-V >= D|Cs]) :-
	generate_var_rels(Var,VDs,Cs).

normal_term_depth(V,(Op_p,C_p)) :-
	term_depth(V,(Op,C)),
	Op = '>', !,
	Op_p = '>=',
	C_p is C + 1.
normal_term_depth(V,(Op,C)) :- term_depth(V,(Op,C)).
	
%% I consider that the size of a variable V is "V > 0".
term_depth(V,('>',0)) :- var(V), !.
term_depth(T,('=',1)) :- atom(T).
term_depth(N,('=',1)) :- number(N).
term_depth(T,(Op,V_p)) :-
	compound(T),
	T =..[_|Ts],
	term_depths(Ts,Sizes),
	worst_size(Sizes,(Op,V)),
	V_p is V + 1.

term_depths([],[]).
term_depths([T|Ts],[S|Ss]) :- term_depth(T,S), term_depths(Ts,Ss).

worst_size(Sizes,(Op,V)) :- worst_size_(Sizes,(Op,V),('=',-1)).

worst_size_([],(Op,V),(Op,V)).
worst_size_([(Op,V)|Ss],(FinalOp,FinalV),(AcOp,AcV)) :-
	(member('>',[Op,AcOp]) -> AcOp_p = '>'
	                        ; AcOp_p = '='),
	ut_max(V,AcV,AcV_p),
	improve_precision((Op,V),(AcOp,AcV),AcV_p,AcV_pp),
	worst_size_(Ss,(FinalOp,FinalV),(AcOp_p,AcV_pp)).

%% If Op is set to '>' and it was an '=' in any of the two, then if
%  the '=' was associated with the max value then we can put '>='
%  instead of '>', which is the same as substracting one.
improve_precision((Op,V),(AcOp,AcV),AcV_p,AcV_pp) :-
	Op \= AcOp,
	((Op,V) = ('=',AcV_p) ; (AcOp,AcV) = ('=',AcV_p)), !,
	AcV_pp is AcV_p - 1.
improve_precision(_,_,AcV_p,AcV_p).


%% Given a term T, the 2nd arg returns a list of pairs (V,D), where V
%  is a variable in T and D its associated depth.
term_vars_and_depths(T,VsAndDs) :- term_vars_(T,VsAndDs,0).
term_vars_(V,[(V,Depth)],Depth) :- var(V), !.
term_vars_([],[],_) :- !.
term_vars_([T|Ts],Vars,D) :- !,
	term_vars_(T,TVs,D),
	term_vars_(Ts,TsVs,D),
	append(TVs,TsVs,Vars).
term_vars_(T,Vars,D) :-
	T =..[_|Args],
	D_p is D+1,
	term_vars_(Args,Vars,D_p).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Auxiliary predicates for the term-size abstraction  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

term_size(T,Size) :-
	term_nfunctors_and_vars(T,NFunctors,Vars,0),
	nfunctors_plus_eachvar(NFunctors,Vars,Size).

term_nfunctors_and_vars(T,N,[T],N) :- var(T), !.
term_nfunctors_and_vars([],N,[],N) :- !.
term_nfunctors_and_vars([T|Ts],N,Vars,NAc) :- !,
	term_nfunctors_and_vars(T,TN,TVars,NAc),
	term_nfunctors_and_vars(Ts,N,TsVars,TN),
	append(TVars,TsVars,Vars).
term_nfunctors_and_vars(T,N,Vars,NAc) :-
	compound(T), !,
	T =..[_|Args],
	NAc_p is NAc + 1,
	term_nfunctors_and_vars(Args,N,Vars,NAc_p).
term_nfunctors_and_vars(_T,N,[],NAc) :- % T is an atom or number
	N is NAc+1.

nfunctors_plus_eachvar(Acu,[],Acu).
nfunctors_plus_eachvar(Acu,[V|Vs],Exp) :-
	nfunctors_plus_eachvar(Acu+V,Vs,Exp).
